import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.peaks import load_pam
from dipy.reconst.shm import descoteaux07_legacy_msg, sph_harm_ind_list
from dipy.sims.voxel import multi_tensor
from dipy.utils.optpkg import optional_package
from dipy.workflows.reconst import (
    ReconstForceFlow,
    ReconstForecastFlow,
    ReconstGQIFlow,
    ReconstPowermapFlow,
    ReconstRUMBAFlow,
    ReconstSFMFlow,
)

_, has_sklearn, _ = optional_package("sklearn")
logging.getLogger().setLevel(logging.INFO)


def simulate_multitensor_data(gtab):
    dwi = np.zeros((2, 2, 2, len(gtab.bvals)))
    # Diffusion of tissue and water compartments are constant for all voxel
    mevals = np.array([[0.0017, 0.0003, 0.0003], [0.003, 0.003, 0.003]])
    # volume fractions
    GTF = np.array([[[0.06, 0.71], [0.33, 0.91]], [[0.0, 0.0], [0.0, 0.0]]])
    for i in range(2):
        for j in range(2):
            gtf = GTF[0, i, j]
            S, p = multi_tensor(
                gtab,
                mevals,
                S0=100,
                angles=[(90, 0), (90, 0)],
                fractions=[(1 - gtf) * 100, gtf * 100],
                snr=None,
            )
            dwi[0, i, j] = S
    return dwi


def test_reconst_rumba():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstRUMBAFlow)


@pytest.mark.skipif(not has_sklearn, reason="Requires sklearn")
def test_reconst_sfm():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstSFMFlow)


def test_reconst_gqi():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstGQIFlow)


def test_reconst_forecast():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstForecastFlow, use_multishell_data=True)


def reconst_flow_core(flow, *, use_multishell_data=None, **kwargs):
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_64D")
        if use_multishell_data:
            bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)

            # FW model requires multishell data
            bvals_2s = np.concatenate((bvals, bvals * 1.5), axis=0)
            bvecs_2s = np.concatenate((bvecs, bvecs), axis=0)
            gtab_2s = gradient_table(bvals_2s, bvecs=bvecs_2s)
            bval_path = Path(out_dir) / os.path.basename(bval_path)
            bvec_path = Path(out_dir) / os.path.basename(bvec_path)
            np.savetxt(bval_path, bvals_2s)
            np.savetxt(bvec_path, bvecs_2s)

            volume = simulate_multitensor_data(gtab_2s)
            data_path = Path(out_dir) / "dwi.nii.gz"
            mask = np.ones_like(volume[..., 0], dtype=bool)
            mask_path = Path(out_dir) / "mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.int32), np.eye(4))
            save_nifti(data_path, volume, np.eye(4))
        else:
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0])
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

        reconst_flow = flow()
        for sh_order_max in [
            8,
        ]:
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                sh_order_max=sh_order_max,
                out_dir=out_dir,
                extract_pam_values=True,
                **kwargs,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            gfa_data = load_nifti_data(gfa_path)
            npt.assert_equal(gfa_data.shape, volume.shape[:-1])

            peaks_dir_path = reconst_flow.last_generated_outputs["out_peaks_dir"]
            peaks_dir_data = load_nifti_data(peaks_dir_path)
            npt.assert_equal(peaks_dir_data.shape[-1], 15)
            npt.assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

            peaks_idx_path = reconst_flow.last_generated_outputs["out_peaks_indices"]
            peaks_idx_data = load_nifti_data(peaks_idx_path)
            npt.assert_equal(peaks_idx_data.shape[-1], 5)
            npt.assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

            peaks_vals_path = reconst_flow.last_generated_outputs["out_peaks_values"]
            peaks_vals_data = load_nifti_data(peaks_vals_path)
            npt.assert_equal(peaks_vals_data.shape[-1], 5)
            npt.assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

            shm_path = reconst_flow.last_generated_outputs["out_shm"]
            shm_data = load_nifti_data(shm_path)
            # Test that the number of coefficients is what you would expect
            # given the order of the sh basis:
            npt.assert_equal(
                shm_data.shape[-1], sph_harm_ind_list(sh_order_max)[0].shape[0]
            )
            npt.assert_equal(shm_data.shape[:-1], volume.shape[:-1])

            pam = load_pam(reconst_flow.last_generated_outputs["out_pam"])
            npt.assert_allclose(
                pam.peak_dirs.reshape(peaks_dir_data.shape), peaks_dir_data
            )

            npt.assert_allclose(pam.peak_values, peaks_vals_data)
            npt.assert_allclose(pam.peak_indices, peaks_idx_data)
            npt.assert_allclose(pam.shm_coeff, shm_data, atol=1e-7)
            npt.assert_allclose(pam.gfa, gfa_data)


def test_reconst_powermap_basic():
    """Test ReconstPowermapFlow basic functionality and parameter variations."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=".*descoteaux07 SH basis.*",
            category=PendingDeprecationWarning,
        )

        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0], dtype=bool)
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

            reconst_flow = ReconstPowermapFlow()

            # Test basic functionality with default parameters
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                out_dir=out_dir,
            )

            powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
            assert os.path.exists(powermap_path), "Powermap file was not created"
            powermap_data = load_nifti_data(powermap_path)
            npt.assert_equal(powermap_data.shape, volume.shape[:-1])
            assert np.all(powermap_data >= 0), "Powermap should be non-negative"
            assert not np.all(powermap_data == 0), "Powermap should not be all zeros"
            assert np.isfinite(
                powermap_data
            ).all(), "Powermap should contain finite values"

            # Test different SH orders
            for sh_order in [4, 6]:
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    sh_order_max=sh_order,
                    out_dir=out_dir,
                    out_powermap=f"powermap_sh{sh_order}.nii.gz",
                )
                powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
                assert os.path.exists(powermap_path)

            # Test different power values and norm factors
            for power in [1, 3]:
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    power=power,
                    out_dir=out_dir,
                    out_powermap=f"powermap_power{power}.nii.gz",
                )
                powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
                assert os.path.exists(powermap_path)

            # Test with smoothing
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                smooth=0.006,
                out_dir=out_dir,
                out_powermap="powermap_smooth.nii.gz",
            )
            powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
            assert os.path.exists(powermap_path)

            # Test non_negative=False
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                non_negative=False,
                out_dir=out_dir,
                out_powermap="powermap_negative_allowed.nii.gz",
            )
            powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
            assert os.path.exists(powermap_path)
            assert reconst_flow.get_short_name() == "powermap"


def test_reconst_powermap_with_shm_files():
    """Test ReconstPowermapFlow with precomputed SH coefficients."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=".*descoteaux07 SH basis.*",
            category=PendingDeprecationWarning,
        )

        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0], dtype=bool)
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

            # Generate SH coefficients using GQI
            reconst_flow = ReconstGQIFlow()
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                sh_order_max=6,
                extract_pam_values=True,
                out_dir=out_dir,
            )

            shm_path = reconst_flow.last_generated_outputs["out_shm"]
            pam_path = reconst_flow.last_generated_outputs["out_pam"]

            # Test with NIfTI SH file
            powermap_flow = ReconstPowermapFlow()
            powermap_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                shm_files=shm_path,
                out_dir=out_dir,
                out_powermap="powermap_from_shm.nii.gz",
            )
            powermap_path = powermap_flow.last_generated_outputs["out_powermap"]
            assert os.path.exists(powermap_path)
            powermap_data = load_nifti_data(powermap_path)
            npt.assert_equal(powermap_data.shape, volume.shape[:-1])

            # Test with PAM file
            powermap_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                shm_files=pam_path,
                out_dir=out_dir,
                out_powermap="powermap_from_pam.nii.gz",
            )
            powermap_path = powermap_flow.last_generated_outputs["out_powermap"]
            assert os.path.exists(powermap_path)


def test_reconst_powermap_edge_cases():
    """Test ReconstPowermapFlow error handling and edge cases."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=".*descoteaux07 SH basis.*",
            category=PendingDeprecationWarning,
        )

        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0], dtype=bool)
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

            reconst_flow = ReconstPowermapFlow()

            # Test with invalid SH file extension
            invalid_shm_path = Path(out_dir) / "invalid.txt"
            invalid_shm_path.write_text("dummy content")
            with pytest.raises(ValueError, match="SH coefficients file must be"):
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    shm_files=str(invalid_shm_path),
                    out_dir=out_dir,
                )

            # Test with invalid sh_basis (should fallback to default)
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                sh_basis="invalid_basis",
                out_dir=out_dir,
                out_powermap="powermap_invalid_basis.nii.gz",
            )
            powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
            assert os.path.exists(powermap_path)

            # Test different norm factors produce different results
            norm_factors = [0.00001, 0.001]
            powermaps = []
            for i, norm_factor in enumerate(norm_factors):
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    norm_factor=norm_factor,
                    out_dir=out_dir,
                    out_powermap=f"powermap_norm{i}.nii.gz",
                )
                powermap_path = reconst_flow.last_generated_outputs["out_powermap"]
                powermap_data = load_nifti_data(powermap_path)
                assert np.all(powermap_data >= 0), "Powermap should be non-negative"
                assert np.isfinite(powermap_data).all(), "All values should be finite"
                powermaps.append(powermap_data)

            # Different norm factors should produce different results
            assert not np.allclose(
                powermaps[0], powermaps[1], rtol=0.1
            ), "Different norm factors should produce different results"


# ---------------------------------------------------------------------------
# FORCE workflow tests
# ---------------------------------------------------------------------------


def _make_fake_simulations(n_gradients, *, n_sims=100):
    """Build a minimal simulations dict for FORCE tests.

    Parameters
    ----------
    n_gradients : int
        Number of DWI gradients (determines signal length).
    n_sims : int
        Number of library entries.

    Returns
    -------
    sims : dict
        Minimal simulations dict accepted by FORCEModel._prepare_library.
    """
    rng = np.random.default_rng(42)
    n_dirs = len(default_sphere.vertices)
    sims = {
        "signals": rng.random((n_sims, n_gradients)).astype(np.float32),
        "fa": rng.random(n_sims).astype(np.float32),
        "md": rng.random(n_sims).astype(np.float32) * 3e-3,
        "rd": rng.random(n_sims).astype(np.float32) * 1e-3,
        "wm_fraction": rng.random(n_sims).astype(np.float32),
        "gm_fraction": rng.random(n_sims).astype(np.float32),
        "csf_fraction": rng.random(n_sims).astype(np.float32),
        "num_fibers": rng.integers(0, 3, n_sims).astype(np.float32),
        "dispersion": rng.random(n_sims).astype(np.float32),
        "nd": rng.random(n_sims).astype(np.float32),
        "labels": rng.integers(0, 2, (n_sims, n_dirs)).astype(np.int8),
        "fraction_array": rng.random((n_sims, 3)).astype(np.float32),
    }
    return sims


def _make_fake_simulations_dki(n_gradients, *, n_sims=100):
    """Like _make_fake_simulations but with DKI fields."""
    sims = _make_fake_simulations(n_gradients, n_sims=n_sims)
    rng = np.random.default_rng(99)
    sims.update(
        {
            "ak": rng.random(n_sims).astype(np.float32),
            "rk": rng.random(n_sims).astype(np.float32),
            "mk": rng.random(n_sims).astype(np.float32),
            "kfa": rng.random(n_sims).astype(np.float32),
        }
    )
    return sims


def _patch_generate(monkeypatch, fake_sims):
    """Patch FORCEModel.generate to inject fake_sims without running simulations."""

    def _fake_generate(self, **kwargs):
        self.simulations = fake_sims
        self._prepare_library()
        return self

    monkeypatch.setattr("dipy.reconst.force.FORCEModel.generate", _fake_generate)


def test_reconst_force_basic(monkeypatch, tmp_path):
    """ReconstForceFlow runs and produces expected output files."""
    data_path, bval_path, bvec_path = get_fnames(name="small_64D")
    volume, affine = load_nifti(data_path)
    mask = np.ones(volume.shape[:3], dtype=np.uint8)
    mask_path = tmp_path / "mask.nii.gz"
    save_nifti(mask_path, mask, affine)

    bvals, _ = read_bvals_bvecs(bval_path, bvec_path)
    n_gradients = len(bvals)

    _patch_generate(monkeypatch, _make_fake_simulations(n_gradients))

    flow = ReconstForceFlow()
    flow.run(
        str(data_path),
        str(bval_path),
        str(bvec_path),
        str(mask_path),
        n_neighbors=5,
        engine="serial",
        out_dir=str(tmp_path),
    )

    for metric in [
        "fa",
        "md",
        "rd",
        "wm_fraction",
        "gm_fraction",
        "csf_fraction",
        "num_fibers",
        "dispersion",
        "nd",
        "uncertainty",
        "ambiguity",
    ]:
        out_path = flow.last_generated_outputs[f"out_{metric}"]
        assert os.path.exists(out_path), f"Missing output: {metric}"
        data_arr = load_nifti_data(out_path)
        npt.assert_equal(data_arr.shape, volume.shape[:3])
        assert np.isfinite(data_arr).all(), f"Non-finite values in {metric}"


def test_reconst_force_select_metrics(monkeypatch, tmp_path):
    """save_metrics limits which output files are written."""
    data_path, bval_path, bvec_path = get_fnames(name="small_64D")
    volume, affine = load_nifti(data_path)
    mask = np.ones(volume.shape[:3], dtype=np.uint8)
    mask_path = tmp_path / "mask.nii.gz"
    save_nifti(mask_path, mask, affine)

    bvals, _ = read_bvals_bvecs(bval_path, bvec_path)
    _patch_generate(monkeypatch, _make_fake_simulations(len(bvals)))

    flow = ReconstForceFlow()
    flow.run(
        str(data_path),
        str(bval_path),
        str(bvec_path),
        str(mask_path),
        n_neighbors=5,
        engine="serial",
        save_metrics=["fa", "md"],
        out_dir=str(tmp_path),
        out_fa="only_fa.nii.gz",
        out_md="only_md.nii.gz",
    )

    assert os.path.exists(flow.last_generated_outputs["out_fa"])
    assert os.path.exists(flow.last_generated_outputs["out_md"])


def test_reconst_force_dki_metrics(monkeypatch, tmp_path):
    """DKI metrics are saved when compute_kurtosis=True."""
    data_path, bval_path, bvec_path = get_fnames(name="small_64D")
    volume, affine = load_nifti(data_path)
    mask = np.ones(volume.shape[:3], dtype=np.uint8)
    mask_path = tmp_path / "mask.nii.gz"
    save_nifti(mask_path, mask, affine)

    bvals, _ = read_bvals_bvecs(bval_path, bvec_path)
    _patch_generate(monkeypatch, _make_fake_simulations_dki(len(bvals)))

    flow = ReconstForceFlow()
    flow.run(
        str(data_path),
        str(bval_path),
        str(bvec_path),
        str(mask_path),
        n_neighbors=5,
        engine="serial",
        compute_kurtosis=True,
        out_dir=str(tmp_path),
    )

    for metric in ["mk", "ak", "rk", "kfa"]:
        out_path = flow.last_generated_outputs[f"out_{metric}"]
        assert os.path.exists(out_path), f"Missing DKI output: {metric}"
        data_arr = load_nifti_data(out_path)
        npt.assert_equal(data_arr.shape, volume.shape[:3])


def test_reconst_force_ray_fallback(monkeypatch, tmp_path):
    """engine='ray' falls back to serial when ray is not available."""
    data_path, bval_path, bvec_path = get_fnames(name="small_64D")
    volume, affine = load_nifti(data_path)
    mask = np.ones(volume.shape[:3], dtype=np.uint8)
    mask_path = tmp_path / "mask.nii.gz"
    save_nifti(mask_path, mask, affine)

    bvals, _ = read_bvals_bvecs(bval_path, bvec_path)
    _patch_generate(monkeypatch, _make_fake_simulations(len(bvals)))

    # Simulate ray being unavailable
    monkeypatch.setattr(
        "dipy.utils.optpkg.optional_package",
        lambda name, *a, **kw: (
            (None, False, None) if name == "ray" else optional_package(name, *a, **kw)
        ),
    )

    flow = ReconstForceFlow()
    flow.run(
        str(data_path),
        str(bval_path),
        str(bvec_path),
        str(mask_path),
        n_neighbors=5,
        engine="ray",
        save_metrics=["fa"],
        out_dir=str(tmp_path),
    )

    out_path = flow.last_generated_outputs["out_fa"]
    assert os.path.exists(out_path)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: cdivision=True
"""
FORCE Core Cython Module

Low-level signal generation for multi-compartment diffusion simulations.
"""

import numpy as np
cimport numpy as cnp

from libc.math cimport exp

ctypedef cnp.float64_t DTYPE_t
ctypedef cnp.uint8_t LABEL_t

# Diffusivity ranges (configurable via set_diffusivity_ranges)
cdef double WM_D_PAR_MIN = 2.0e-3
cdef double WM_D_PAR_MAX = 3.0e-3
cdef double WM_D_PERP_MIN = 0.3e-3
cdef double WM_D_PERP_MAX = 1.5e-3
cdef double GM_D_MIN = 0.7e-3
cdef double GM_D_MAX = 1.2e-3
cdef double CSF_D = 3.0e-3


def set_diffusivity_ranges(
    wm_d_par_range=(2.0e-3, 3.0e-3),
    wm_d_perp_range=(0.3e-3, 1.5e-3),
    gm_d_iso_range=(0.7e-3, 1.2e-3),
    csf_d=3.0e-3,
):
    """
    Update the diffusivity ranges used by the simulators.

    Parameters
    ----------
    wm_d_par_range : tuple or float
        White matter parallel diffusivity range (min, max) in mm^2/s.
        If a single float, uses fixed value.
    wm_d_perp_range : tuple or float
        White matter perpendicular diffusivity range in mm^2/s.
    gm_d_iso_range : tuple or float
        Gray matter isotropic diffusivity range in mm^2/s.
    csf_d : float
        CSF isotropic diffusivity in mm^2/s.
    """
    global WM_D_PAR_MIN, WM_D_PAR_MAX, WM_D_PERP_MIN, WM_D_PERP_MAX
    global GM_D_MIN, GM_D_MAX, CSF_D

    def _validate_val_or_pair(name, val):
        try:
            if hasattr(val, "__len__") and len(val) == 2:
                lo, hi = float(val[0]), float(val[1])
            else:
                lo = hi = float(val)
        except Exception as e:
            raise ValueError(f"{name} must be a float or (min, max) pair") from e
        if lo > hi:
            raise ValueError(f"{name} min must be <= max")
        return lo, hi

    WM_D_PAR_MIN, WM_D_PAR_MAX = _validate_val_or_pair("wm_d_par_range", wm_d_par_range)
    WM_D_PERP_MIN, WM_D_PERP_MAX = _validate_val_or_pair("wm_d_perp_range", wm_d_perp_range)
    GM_D_MIN, GM_D_MAX = _validate_val_or_pair("gm_d_iso_range", gm_d_iso_range)
    CSF_D = float(csf_d)


cdef inline double get_dperp_extra(double d_par, double f_intra) noexcept:
    """
    Compute extra-axonal perpendicular diffusivity using tortuosity model.

    Parameters
    ----------
    d_par : double
        Parallel diffusivity.
    f_intra : double
        Intra-axonal volume fraction.

    Returns
    -------
    d_perp : double
        Extra-axonal perpendicular diffusivity.
    """
    return d_par * (1.0 - f_intra) / (1.0 + f_intra)


cdef inline double fa_stick_zeppelin(double d_par, double d_perp, double f_intra) noexcept:
    """
    Compute microFA for stick-zeppelin model.

    Calculates microscopic fractional anisotropy for a mixture of
    sticks (Da, 0, 0) and zeppelins (Da, dperp_ex, dperp_ex).

    Parameters
    ----------
    d_par : double
        Parallel diffusivity.
    d_perp : double
        Perpendicular diffusivity.
    f_intra : double
        Intra-axonal volume fraction.

    Returns
    -------
    ufa : double
        Microscopic fractional anisotropy.
    """
    cdef double e = 1.0 - f_intra
    cdef double num = d_par - e * d_perp
    cdef double den = (d_par * d_par + 2.0 * (e * e) * d_perp * d_perp) ** 0.5
    return num / den


cdef inline double _sample_uniform_or_fixed(double lo, double hi) noexcept:
    """Sample uniformly from [lo, hi] or return fixed value if lo == hi."""
    if lo == hi:
        return lo
    else:
        return float(np.random.uniform(lo, hi))


cdef inline double sample_wm_d_par() noexcept:
    """Sample white matter parallel diffusivity."""
    return _sample_uniform_or_fixed(WM_D_PAR_MIN, WM_D_PAR_MAX)


cdef inline double sample_wm_d_perp() noexcept:
    """Sample white matter perpendicular diffusivity."""
    return _sample_uniform_or_fixed(WM_D_PERP_MIN, WM_D_PERP_MAX)


cdef inline double sample_gm_d_iso() noexcept:
    """Sample gray matter isotropic diffusivity."""
    return _sample_uniform_or_fixed(GM_D_MIN, GM_D_MAX)


cdef Py_ssize_t _closest_direction(
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere,
    cnp.ndarray[DTYPE_t, ndim=1] vec
) noexcept:
    """
    Find the closest direction on a sphere to a given vector.

    Parameters
    ----------
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.
    vec : ndarray (3,)
        Target direction vector.

    Returns
    -------
    idx : int
        Index of the closest direction.
    """
    cdef Py_ssize_t i, n = target_sphere.shape[0]
    cdef double best = 1e300
    cdef double dist, dx, dy, dz
    cdef Py_ssize_t best_idx = 0
    cdef const double[:, :] ts_mv = target_sphere

    for i in range(n):
        dx = ts_mv[i, 0] - vec[0]
        dy = ts_mv[i, 1] - vec[1]
        dz = ts_mv[i, 2] - vec[2]
        dist = dx * dx + dy * dy + dz * dz
        if dist < best:
            best = dist
            best_idx = i

    return best_idx


def angle_between(v1, v2):
    """
    Compute the angle between two vectors.

    Parameters
    ----------
    v1 : ndarray (3,)
        First vector.
    v2 : ndarray (3,)
        Second vector.

    Returns
    -------
    angle : float
        Angle between the two vectors in radians.
    """
    cos_theta = np.clip(np.dot(v1, v2), -1.0, 1.0)
    return np.arccos(cos_theta)


def is_angle_valid(angle, *, threshold=30):
    """
    Check if an angle satisfies minimum separation constraint.

    Parameters
    ----------
    angle : float
        Angle to check in radians.
    threshold : float, optional
        Minimum separation threshold in degrees.

    Returns
    -------
    is_valid : bool
        True if the angle satisfies the separation constraint.
    """
    angle_degrees = np.degrees(angle)
    return not (
        (-1 * threshold <= angle_degrees <= threshold) or
        (180 - threshold <= angle_degrees <= 180 + threshold)
    )


cpdef tuple generate_single_fiber(
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere,
    cnp.ndarray[DTYPE_t, ndim=3] evecs,
    object bingham_sf,
    cnp.ndarray[DTYPE_t, ndim=1] odi_list,
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] bvecs,
    object multi_tensor_func,
    bint tortuosity
):
    """
    Generate diffusion signal for a single fiber population.

    Parameters
    ----------
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.
    evecs : ndarray (N, 3, 3)
        Eigenvector matrices for each sphere direction.
    bingham_sf : dict
        Pre-computed Bingham spherical functions.
    odi_list : ndarray
        List of orientation dispersion index values.
    bvals : ndarray
        B-values.
    bvecs : ndarray (M, 3)
        Gradient directions.
    multi_tensor_func : callable
        Function to compute multi-tensor signal.
    tortuosity : bool
        Whether to use tortuosity constraint.

    Returns
    -------
    tuple
        (signal, labels, num_fibers, dispersion, placeholder,
         neurite_density, odf, d_par, d_perp, fractions, f_ins)
    """
    cdef:
        Py_ssize_t n_dirs = target_sphere.shape[0]
        double f_intra = float(np.random.uniform(0.6, 0.9))
        double f_extra = 1.0 - f_intra
        double d_par = sample_wm_d_par()
        double d_perp_extra
        double S0 = 100.0
        int idx

    if tortuosity:
        d_perp_extra = get_dperp_extra(d_par, f_intra)
    else:
        d_perp_extra = sample_wm_d_perp()

    labels = np.zeros(n_dirs, dtype=np.uint8)

    mevals_ex = np.zeros_like(target_sphere, dtype=np.float64)
    mevals_ex[:, 0] = d_par
    mevals_ex[:, 1] = d_perp_extra
    mevals_ex[:, 2] = d_perp_extra

    mevals_in = np.zeros_like(target_sphere, dtype=np.float64)
    mevals_in[:, 0] = d_par
    mevals_in[:, 1] = 0.0
    mevals_in[:, 2] = 0.0

    idx = int(np.random.randint(0, n_dirs))
    true_stick = target_sphere[idx]

    factor = float(np.random.choice(odi_list))

    fodf_gt = bingham_sf[idx][factor]
    fodf_gt = np.ascontiguousarray(fodf_gt, dtype=np.float64)
    fodf_gt = fodf_gt / np.sum(fodf_gt)

    S_in = multi_tensor_func(mevals_in, evecs, fodf_gt * 100.0, bvals, bvecs)
    S_ex = multi_tensor_func(mevals_ex, evecs, fodf_gt * 100.0, bvals, bvecs)

    S = f_intra * S_in + f_extra * S_ex

    nearest = _closest_direction(target_sphere, true_stick)
    labels[nearest] = 1

    return (
        S * S0,
        labels,
        1,
        factor,
        0.0,
        1.0 * f_intra,
        fodf_gt,
        d_par,
        f_extra * d_perp_extra,
        [1.0, 0.0, 0.0],
        [f_intra],
    )


cpdef tuple generate_two_fibers(
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere,
    cnp.ndarray[DTYPE_t, ndim=3] evecs,
    object bingham_sf,
    cnp.ndarray[DTYPE_t, ndim=1] odi_list,
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] bvecs,
    object multi_tensor_func,
    bint tortuosity
):
    """
    Generate diffusion signal for two crossing fiber populations.

    Parameters
    ----------
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.
    evecs : ndarray (N, 3, 3)
        Eigenvector matrices for each sphere direction.
    bingham_sf : dict
        Pre-computed Bingham spherical functions.
    odi_list : ndarray
        List of orientation dispersion index values.
    bvals : ndarray
        B-values.
    bvecs : ndarray (M, 3)
        Gradient directions.
    multi_tensor_func : callable
        Function to compute multi-tensor signal.
    tortuosity : bool
        Whether to use tortuosity constraint.

    Returns
    -------
    tuple
        Signal and associated parameters.
    """
    cdef:
        Py_ssize_t n_dirs = target_sphere.shape[0]
        double S0 = 100.0
        double d_par, d_perp_extra
        double fiber_frac1
        double wm_nd, wm_d_perp
        int i
        int idx0, idx1

    f_in = np.random.uniform(0.6, 0.9, 2).astype(np.float64)
    fiber_frac1 = float(np.random.uniform(0.2, 0.8))
    fiber_fractions = [fiber_frac1, 1.0 - fiber_frac1]

    d_par = sample_wm_d_par()
    if tortuosity:
        d_perp_extra = get_dperp_extra(d_par, float(f_in[0]))
    else:
        d_perp_extra = sample_wm_d_perp()

    labels = np.zeros(n_dirs, dtype=np.uint8)

    mevals_ex = np.zeros_like(target_sphere, dtype=np.float64)
    mevals_ex[:, 0] = d_par
    mevals_ex[:, 1] = d_perp_extra
    mevals_ex[:, 2] = d_perp_extra

    mevals_in = np.zeros_like(target_sphere, dtype=np.float64)
    mevals_in[:, 0] = d_par
    mevals_in[:, 1] = 0.0
    mevals_in[:, 2] = 0.0

    _MAX_ANGLE_TRIES = 10000
    index = np.random.randint(0, n_dirs, 2)
    _angle_tries = 0
    while not is_angle_valid(
        angle_between(target_sphere[index[0]], target_sphere[index[1]])
    ):
        index = np.random.randint(0, n_dirs, 2)
        _angle_tries += 1
        if _angle_tries >= _MAX_ANGLE_TRIES:
            raise RuntimeError(
                "Could not find two fiber directions with sufficient angular "
                "separation after %d attempts. Consider using a finer sphere."
                % _MAX_ANGLE_TRIES
            )

    idx0 = int(index[0])
    idx1 = int(index[1])
    true_stick1 = target_sphere[idx0]
    true_stick2 = target_sphere[idx1]

    factor = float(np.random.choice(odi_list))

    S = np.zeros(bvals.shape[0], dtype=np.float64)
    fodf = np.zeros(n_dirs, dtype=np.float64)

    for i in range(2):
        fodf_gt = bingham_sf[int(index[i])][factor]
        fodf_gt = np.ascontiguousarray(fodf_gt, dtype=np.float64)
        fodf_gt = fodf_gt / np.sum(fodf_gt)

        fodf += fiber_fractions[i] * fodf_gt

        S_in = multi_tensor_func(mevals_in, evecs, fodf_gt * 100.0, bvals, bvecs)
        S_ex = multi_tensor_func(mevals_ex, evecs, fodf_gt * 100.0, bvals, bvecs)

        f_intra = float(f_in[i])
        f_extra = 1.0 - f_intra

        S += fiber_fractions[i] * (f_intra * S_in + f_extra * S_ex)

    labels[_closest_direction(target_sphere, true_stick1)] = 1
    labels[_closest_direction(target_sphere, true_stick2)] = 1

    wm_nd = fiber_fractions[0] * float(f_in[0]) + fiber_fractions[1] * float(f_in[1])
    wm_d_perp = (
        fiber_fractions[0] * (1.0 - float(f_in[0])) * d_perp_extra
        + fiber_fractions[1] * (1.0 - float(f_in[1])) * d_perp_extra
    )

    return (
        S * S0,
        labels,
        2,
        factor,
        0.0,
        wm_nd,
        fodf,
        d_par,
        wm_d_perp,
        [fiber_frac1, 1.0 - fiber_frac1, 0.0],
        f_in.tolist(),
    )


cpdef tuple generate_three_fibers(
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere,
    cnp.ndarray[DTYPE_t, ndim=3] evecs,
    object bingham_sf,
    cnp.ndarray[DTYPE_t, ndim=1] odi_list,
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] bvecs,
    object multi_tensor_func,
    bint tortuosity
):
    """
    Generate diffusion signal for three crossing fiber populations.

    Parameters
    ----------
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.
    evecs : ndarray (N, 3, 3)
        Eigenvector matrices for each sphere direction.
    bingham_sf : dict
        Pre-computed Bingham spherical functions.
    odi_list : ndarray
        List of orientation dispersion index values.
    bvals : ndarray
        B-values.
    bvecs : ndarray (M, 3)
        Gradient directions.
    multi_tensor_func : callable
        Function to compute multi-tensor signal.
    tortuosity : bool
        Whether to use tortuosity constraint.

    Returns
    -------
    tuple
        Signal and associated parameters.
    """
    cdef:
        Py_ssize_t n_dirs = target_sphere.shape[0]
        double S0 = 100.0
        double d_par, d_perp_extra
        double wm_nd = 0.0
        double wm_d_perp = 0.0
        int k
        int idx0, idx1, idx2

    f_in = np.random.uniform(0.6, 0.9, 3).astype(np.float64)
    fiber_fracs = np.random.dirichlet([1.0, 1.0, 1.0]).astype(np.float64)
    while np.any(fiber_fracs < 0.2):
        fiber_fracs = np.random.dirichlet([1.0, 1.0, 1.0]).astype(np.float64)

    d_par = sample_wm_d_par()
    if tortuosity:
        d_perp_extra = get_dperp_extra(d_par, float(f_in[0]))
    else:
        d_perp_extra = sample_wm_d_perp()

    labels = np.zeros(n_dirs, dtype=np.uint8)

    mevals_ex = np.zeros_like(target_sphere, dtype=np.float64)
    mevals_ex[:, 0] = d_par
    mevals_ex[:, 1] = d_perp_extra
    mevals_ex[:, 2] = d_perp_extra

    mevals_in = np.zeros_like(target_sphere, dtype=np.float64)
    mevals_in[:, 0] = d_par
    mevals_in[:, 1] = 0.0
    mevals_in[:, 2] = 0.0

    _MAX_ANGLE_TRIES = 10000
    index = np.random.randint(0, n_dirs, 3)
    _angle_tries = 0
    while (
        not is_angle_valid(
            angle_between(target_sphere[index[0]], target_sphere[index[1]]),
            threshold=60
        )
        or not is_angle_valid(
            angle_between(target_sphere[index[0]], target_sphere[index[2]]),
            threshold=60
        )
        or not is_angle_valid(
            angle_between(target_sphere[index[1]], target_sphere[index[2]]),
            threshold=60
        )
    ):
        index = np.random.randint(0, n_dirs, 3)
        _angle_tries += 1
        if _angle_tries >= _MAX_ANGLE_TRIES:
            raise RuntimeError(
                "Could not find three fiber directions with sufficient angular "
                "separation after %d attempts. Consider using a finer sphere."
                % _MAX_ANGLE_TRIES
            )

    idx0 = int(index[0])
    idx1 = int(index[1])
    idx2 = int(index[2])

    true_stick1 = target_sphere[idx0]
    true_stick2 = target_sphere[idx1]
    true_stick3 = target_sphere[idx2]

    factor = float(np.random.choice(odi_list))

    fodf = np.zeros(n_dirs, dtype=np.float64)
    S = np.zeros(bvals.shape[0], dtype=np.float64)

    for k in range(3):
        fodf_gt = bingham_sf[int(index[k])][factor]
        fodf_gt = np.ascontiguousarray(fodf_gt, dtype=np.float64)
        fodf_gt = fodf_gt / np.sum(fodf_gt)

        fodf += fiber_fracs[k] * fodf_gt

        S_in = multi_tensor_func(mevals_in, evecs, fodf_gt * 100.0, bvals, bvecs)
        S_ex = multi_tensor_func(mevals_ex, evecs, fodf_gt * 100.0, bvals, bvecs)

        S += fiber_fracs[k] * (float(f_in[k]) * S_in + (1.0 - float(f_in[k])) * S_ex)

    labels[_closest_direction(target_sphere, true_stick1)] = 1
    labels[_closest_direction(target_sphere, true_stick2)] = 1
    labels[_closest_direction(target_sphere, true_stick3)] = 1

    for k in range(3):
        wm_nd += fiber_fracs[k] * float(f_in[k])
        wm_d_perp += fiber_fracs[k] * (1.0 - float(f_in[k])) * d_perp_extra

    return (
        S * S0,
        labels,
        3,
        factor,
        0.0,
        wm_nd,
        fodf,
        d_par,
        wm_d_perp,
        fiber_fracs.tolist(),
        f_in.tolist(),
    )


cpdef tuple create_wm_signal(
    int num_fib,
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere,
    cnp.ndarray[DTYPE_t, ndim=3] evecs,
    object bingham_sf,
    cnp.ndarray[DTYPE_t, ndim=1] odi_list,
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] bvecs,
    object multi_tensor_func,
    bint tortuosity
):
    """
    Create white matter signal with specified number of fibers.

    Parameters
    ----------
    num_fib : int
        Number of fiber populations (1, 2, or 3).
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.
    evecs : ndarray (N, 3, 3)
        Eigenvector matrices for each sphere direction.
    bingham_sf : dict
        Pre-computed Bingham spherical functions.
    odi_list : ndarray
        List of orientation dispersion index values.
    bvals : ndarray
        B-values.
    bvecs : ndarray (M, 3)
        Gradient directions.
    multi_tensor_func : callable
        Function to compute multi-tensor signal.
    tortuosity : bool
        Whether to use tortuosity constraint.

    Returns
    -------
    tuple
        Signal and associated parameters.
    """
    if num_fib == 1:
        return generate_single_fiber(
            target_sphere, evecs, bingham_sf, odi_list,
            bvals, bvecs, multi_tensor_func, tortuosity
        )
    elif num_fib == 2:
        return generate_two_fibers(
            target_sphere, evecs, bingham_sf, odi_list,
            bvals, bvecs, multi_tensor_func, tortuosity
        )
    elif num_fib == 3:
        return generate_three_fibers(
            target_sphere, evecs, bingham_sf, odi_list,
            bvals, bvecs, multi_tensor_func, tortuosity
        )
    else:
        raise ValueError("num_fib must be 1, 2 or 3")


cpdef tuple create_gm_signal(
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere
):
    """
    Create gray matter isotropic diffusion signal.

    Parameters
    ----------
    bvals : ndarray
        B-values.
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.

    Returns
    -------
    tuple
        (signal, labels, num_fibers, dispersion, placeholder,
         neurite_density, odf, d_par, d_perp)
    """
    cdef:
        Py_ssize_t n_dirs = target_sphere.shape[0]
        double d = sample_gm_d_iso()

    signal = np.exp(-bvals * d) * 100.0
    labels = np.zeros(n_dirs, dtype=np.uint8)
    gm_odf = np.ones(n_dirs, dtype=np.float64) / float(n_dirs)

    return (
        signal,
        labels,
        0,
        1.0,
        0.0,
        0.0,
        gm_odf,
        d,
        d,
    )


cpdef tuple create_csf_signal(
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere
):
    """
    Create CSF free water diffusion signal.

    Parameters
    ----------
    bvals : ndarray
        B-values.
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.

    Returns
    -------
    tuple
        (signal, labels, num_fibers, placeholder, free_water_fraction,
         placeholder, odf, d_par, d_perp)
    """
    cdef:
        Py_ssize_t n_dirs = target_sphere.shape[0]
        double d = CSF_D

    signal = np.exp(-bvals * d) * 100.0
    labels = np.zeros(n_dirs, dtype=np.uint8)
    csf_odf = np.zeros(n_dirs, dtype=np.float64)

    return (
        signal,
        labels,
        0,
        1.0,
        1.0,
        0.0,
        csf_odf,
        d,
        d,
    )


cpdef tuple create_mixed_signal(
    cnp.ndarray[DTYPE_t, ndim=2] target_sphere,
    cnp.ndarray[DTYPE_t, ndim=3] evecs,
    object bingham_sf,
    cnp.ndarray[DTYPE_t, ndim=1] odi_list,
    cnp.ndarray[DTYPE_t, ndim=1] bvals,
    cnp.ndarray[DTYPE_t, ndim=2] bvecs,
    object multi_tensor_func,
    double wm_threshold,
    bint tortuosity
):
    """
    Create mixed WM/GM/CSF tissue signal.

    This is the main simulation function that generates realistic
    voxel signals with mixed tissue contributions.

    Parameters
    ----------
    target_sphere : ndarray (N, 3)
        Unit sphere vertices.
    evecs : ndarray (N, 3, 3)
        Eigenvector matrices for each sphere direction.
    bingham_sf : dict
        Pre-computed Bingham spherical functions.
    odi_list : ndarray
        List of orientation dispersion index values.
    bvals : ndarray
        B-values.
    bvecs : ndarray (M, 3)
        Gradient directions.
    multi_tensor_func : callable
        Function to compute multi-tensor signal.
    wm_threshold : float
        Minimum WM fraction to include fiber labels.
    tortuosity : bool
        Whether to use tortuosity constraint.

    Returns
    -------
    tuple
        (signal, labels, num_fibers, dispersion, wm_fraction,
         gm_fraction, csf_fraction, neurite_density, odf,
         ufa_wm, ufa_voxel, fiber_fractions, wm_disp,
         wm_d_par, wm_d_perp, gm_d_par, csf_d_par, f_ins)
    """
    cdef:
        Py_ssize_t n_dirs = target_sphere.shape[0]
        double wm_fraction
        double gm_fraction
        double csf_fraction
        int num_fiber
        double odi
        double nd
        double ufa_wm = 0.0
        double ufa_voxel
        int k

    fractions = np.random.dirichlet([2.0, 1.0, 1.0]).astype(np.float64)
    wm_fraction = float(fractions[0])
    gm_fraction = float(fractions[1])
    csf_fraction = float(fractions[2])

    num_fiber = int(np.random.choice([1, 2, 3], p=[0.1, 0.2, 0.7]))

    wm_result = create_wm_signal(
        num_fiber,
        target_sphere,
        evecs,
        bingham_sf,
        odi_list,
        bvals,
        bvecs,
        multi_tensor_func,
        tortuosity,
    )
    wm_signal = wm_result[0]
    wm_label = wm_result[1]
    wm_num_fib = wm_result[2]
    wm_disp = wm_result[3]
    wm_nd = wm_result[5]
    wm_odf = wm_result[6]
    wm_d_par = wm_result[7]
    wm_d_perp = wm_result[8]
    fracs = wm_result[9]
    f_ins = wm_result[10]

    gm_result = create_gm_signal(bvals, target_sphere)
    gm_signal = gm_result[0]
    gm_disp = gm_result[3]
    gm_nd = gm_result[5]
    gm_d_par = gm_result[7]

    csf_result = create_csf_signal(bvals, target_sphere)
    csf_signal = csf_result[0]
    csf_d_par = csf_result[7]

    odi = wm_fraction * float(wm_disp) + gm_fraction * float(gm_disp) + csf_fraction * 1.0
    nd = wm_fraction * float(wm_nd) + gm_fraction * float(gm_nd)

    combined_signal = (
        wm_fraction * wm_signal
        + gm_fraction * gm_signal
        + csf_fraction * csf_signal
    )

    if wm_fraction > wm_threshold:
        combined_odf = 50.0 * wm_fraction * wm_odf
    else:
        wm_label = np.zeros(n_dirs, dtype=np.uint8)
        combined_odf = np.zeros(n_dirs, dtype=np.float16)

    for k in range(wm_num_fib):
        ufa_wm += fa_stick_zeppelin(
            float(wm_d_par),
            float(wm_d_perp),
            float(f_ins[k]),
        ) * float(fracs[k])

    ufa_voxel = ufa_wm * wm_fraction

    frac_arr = np.zeros(3, dtype=np.float32)
    for k in range(min(3, len(fracs))):
        frac_arr[k] = float(fracs[k])

    f_ins_arr = np.zeros(3, dtype=np.float32)
    for k in range(min(3, len(f_ins))):
        f_ins_arr[k] = float(f_ins[k])

    return (
        combined_signal,
        wm_label,
        wm_num_fib,
        odi,
        wm_fraction,
        gm_fraction,
        csf_fraction,
        nd,
        combined_odf.astype(np.float16),
        float(ufa_wm),
        float(ufa_voxel),
        frac_arr,
        float(wm_disp),
        float(wm_d_par),
        float(wm_d_perp),
        float(gm_d_par),
        float(csf_d_par),
        f_ins_arr,
    )



from libc.stdint cimport int64_t

cdef void select_top_k_parallel(
    const float* distances,
    size_t n_queries,
    size_t n_database,
    size_t k,
    float* out_distances,
    int64_t* out_indices
) noexcept nogil

cdef void heap_init_parallel(
    size_t n_queries,
    size_t k,
    float* out_distances,
    int64_t* out_indices
) noexcept nogil

cdef void heap_update_batch_parallel(
    const float* distances,
    size_t n_queries,
    size_t chunk_size,
    size_t chunk_offset,
    size_t k,
    float* out_distances,
    int64_t* out_indices
) noexcept nogil

cdef void heap_finalize_parallel(
    size_t n_queries,
    size_t k,
    float* out_distances,
    int64_t* out_indices
) noexcept nogil



cimport numpy as np
cimport numpy as cnp

cdef class PmfGen:
    cdef:
        double[:] pmf
        double[:, :, :, :] data
        double[:, :] vertices
        object sphere

    cdef double* get_pmf_c(self, double* point, double* out) noexcept nogil
    cdef int find_closest(self, double* xyz) noexcept nogil
    cdef double get_pmf_value_c(self, double* point, double* xyz) noexcept nogil
    cdef cnp.npy_intp get_peaks_c(self,
                                  double* point,
                                  double* out_values,
                                  double* out_indices,
                                  double* out_weights,
                                  cnp.npy_intp max_peaks,
                                  cnp.npy_intp* out_valid) noexcept nogil
    pass


cdef class SimplePmfGen(PmfGen):
    pass


cdef class SHCoeffPmfGen(PmfGen):
    cdef:
        double[:, :] B
        double[:] coeff
    pass


cdef class SimplePeakGen(PmfGen):
    cdef:
        cnp.int32_t[:, :, :, :] peak_indices
        double[:, :, :, :] peak_values
        cnp.int32_t* peak_indices_ptr
        double* peak_values_ptr
        int max_peaks
        cnp.npy_intp[4] peak_shape
        cnp.npy_intp[4] peak_indices_strides
        cnp.npy_intp[4] peak_values_strides
    cdef void _compute_trilinear(self,
                                 double* point,
                                 double* out_weights,
                                 cnp.npy_intp* out_index) noexcept nogil
    cdef cnp.npy_intp _inside_global_bounds(self,
                                            cnp.npy_intp* index) noexcept nogil
    pass



import numpy as np
import numpy.testing as npt

import dipy.utils.parallel as para
from dipy.utils.parallel import _available_ram, auto_ray_chunk_size


def power_it(num, n=2):
    # We define a function of the right form for parallelization
    return num**n


ENGINES = ["serial"]
if para.has_dask:
    ENGINES.append("dask")
if para.has_joblib:
    ENGINES.append("joblib")
if para.has_ray:
    ENGINES.append("ray")


def test_paramap():
    my_array = np.arange(100).reshape(10, 10)
    my_list = list(my_array.ravel())
    for engine in ENGINES:
        for backend in ["threading", "multiprocessing"]:
            npt.assert_array_equal(
                para.paramap(
                    power_it,
                    my_list,
                    engine=engine,
                    backend=backend,
                    out_shape=my_array.shape,
                ),
                power_it(my_array),
            )

            # If it's not reshaped, the first item should be the item 0, 0:
            npt.assert_equal(
                para.paramap(power_it, my_list, engine=engine, backend=backend)[0],
                power_it(my_array[0, 0]),
            )


def test_paramap_sequence_kwargs():
    my_array = np.arange(10)
    kwargs_sequence = [{"n": i} for i in range(len(my_array))]
    my_list = list(my_array.ravel())
    for engine in ENGINES:
        for backend in ["threading", "multiprocessing"]:
            npt.assert_array_equal(
                para.paramap(
                    power_it,
                    my_list,
                    engine=engine,
                    backend=backend,
                    out_shape=my_array.shape,
                    func_kwargs=kwargs_sequence,
                ),
                [
                    power_it(ii, **kwargs)
                    for ii, kwargs in zip(my_array, kwargs_sequence)
                ],
            )


def test_available_ram():
    ram = _available_ram()
    assert isinstance(ram, int)
    assert ram > 0
    # Sanity: at least 64 MB, at most 64 TiB
    assert 64 * 1024**2 <= ram <= 64 * 1024**4


def test_auto_ray_chunk_size_returns_positive_int():
    chunk = auto_ray_chunk_size(n_jobs=4, n_gradients=160)
    assert isinstance(chunk, int)
    assert chunk > 0


def test_auto_ray_chunk_size_efficiency_floor():
    # Even with zero RAM budget the result must not fall below N_min
    from dipy.utils.parallel import _RAY_TASK_OVERHEAD_MS, _RAY_T_VOXEL_MS

    n_min = int(_RAY_TASK_OVERHEAD_MS / (0.10 * _RAY_T_VOXEL_MS))
    chunk = auto_ray_chunk_size(
        n_jobs=8,
        n_gradients=160,
        shared_obj_nbytes=10 * 1024**4,  # 10 TiB — exhausts any RAM budget
    )
    assert chunk >= n_min


def test_auto_ray_chunk_size_parallelism_cap():
    # With n_vox given, chunk <= n_vox // n_jobs  (enough chunks for all workers)
    n_vox, n_jobs = 10_000, 4
    chunk = auto_ray_chunk_size(n_jobs=n_jobs, n_gradients=160, n_vox=n_vox)
    assert chunk <= n_vox // n_jobs


def test_auto_ray_chunk_size_memory_scales_with_ram():
    # More RAM budget  →  larger (or equal) chunk when n_par is not binding
    large_n_vox = 50_000_000  # large enough that n_par does not cap
    chunk_tight = auto_ray_chunk_size(
        n_jobs=8,
        n_gradients=160,
        n_vox=large_n_vox,
        shared_obj_nbytes=0,
        ram_fraction=0.001,  # very small budget
    )
    chunk_generous = auto_ray_chunk_size(
        n_jobs=8,
        n_gradients=160,
        n_vox=large_n_vox,
        shared_obj_nbytes=0,
        ram_fraction=0.9,  # large budget
    )
    assert chunk_tight <= chunk_generous


def test_auto_ray_chunk_size_negative_n_jobs():
    # n_jobs=-1 must not collapse to 1 worker (which would yield 1 giant chunk)
    from dipy.utils.multiproc import determine_num_processes

    n_vox = 100_000
    chunk_neg = auto_ray_chunk_size(n_jobs=-1, n_gradients=160, n_vox=n_vox)
    chunk_pos = auto_ray_chunk_size(
        n_jobs=determine_num_processes(-1), n_gradients=160, n_vox=n_vox
    )
    assert chunk_neg == chunk_pos
    # Must produce more than one chunk
    assert chunk_neg < n_vox


def test_auto_ray_chunk_size_more_jobs_smaller_chunks():
    # More workers → smaller per-worker budget → smaller chunks
    n_vox = 1_000_000
    chunk_few = auto_ray_chunk_size(n_jobs=2, n_gradients=160, n_vox=n_vox)
    chunk_many = auto_ray_chunk_size(n_jobs=32, n_gradients=160, n_vox=n_vox)
    assert chunk_few >= chunk_many



from collections.abc import Sequence
import json
import multiprocessing
import shutil
import tempfile

import numpy as np
from tqdm.auto import tqdm

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.multiproc import determine_num_processes
from dipy.utils.optpkg import optional_package

ray, has_ray, _ = optional_package("ray")
joblib, has_joblib, _ = optional_package("joblib")
dask, has_dask, _ = optional_package("dask")
_RAY_TASK_OVERHEAD_MS = 3.0  # empirical fixed cost per Ray task (ms)
_RAY_T_VOXEL_MS = 0.015  # conservative compute time per voxel (ms)


def _available_ram():
    """Estimate available RAM in bytes using stdlib only.

    * **Linux** — reads ``MemAvailable`` from ``/proc/meminfo`` (accurate).
    * **Windows** — queries ``GlobalMemoryStatusEx`` via ``ctypes`` (accurate).
    * **macOS / other POSIX** — returns half of total physical RAM via
      ``os.sysconf`` (conservative approximation).
    * **Fallback** — 4 GiB when none of the above succeed.
    """
    import platform

    system = platform.system()
    try:
        if system == "Linux":
            with open("/proc/meminfo") as fh:
                for line in fh:
                    if line.startswith("MemAvailable:"):
                        return int(line.split()[1]) * 1024
        elif system == "Windows":
            import ctypes

            class _MEMORYSTATUSEX(ctypes.Structure):
                _fields_ = [
                    ("dwLength", ctypes.c_ulong),
                    ("dwMemoryLoad", ctypes.c_ulong),
                    ("ullTotalPhys", ctypes.c_ulonglong),
                    ("ullAvailPhys", ctypes.c_ulonglong),
                    ("ullTotalPageFile", ctypes.c_ulonglong),
                    ("ullAvailPageFile", ctypes.c_ulonglong),
                    ("ullTotalVirtual", ctypes.c_ulonglong),
                    ("ullAvailVirtual", ctypes.c_ulonglong),
                    ("ullAvailExtendedVirtual", ctypes.c_ulonglong),
                ]

            mem = _MEMORYSTATUSEX()
            mem.dwLength = ctypes.sizeof(_MEMORYSTATUSEX)
            ctypes.windll.kernel32.GlobalMemoryStatusEx(ctypes.byref(mem))
            return mem.ullAvailPhys
        else:
            import os

            pages = os.sysconf("SC_PHYS_PAGES")
            page_size = os.sysconf("SC_PAGE_SIZE")
            return (pages * page_size) // 2
    except Exception:
        pass
    return 4 * 1024**3


def auto_ray_chunk_size(
    *,
    n_jobs,
    n_gradients,
    n_vox=None,
    shared_obj_nbytes=0,
    ram_fraction=0.4,
    min_efficiency=0.10,
    output_bytes_per_voxel=4096,
):
    """Compute a memory- and efficiency-aware Ray chunk size.

    Three constraints are balanced:

    * **Efficiency lower bound** — chunk large enough that Ray task overhead
      stays below ``min_efficiency`` of total compute time::

          N_min = ray_overhead_ms / (min_efficiency × t_voxel_ms)
                ≈ 3 ms / (0.10 × 0.015 ms) ≈ 2 000 voxels

    * **Memory upper bound** — all ``n_jobs`` in-flight chunks fit in the
      available RAM budget::

          N_mem = (available_RAM × ram_fraction − shared_obj_bytes)
                  ────────────────────────────────────────────────────
                          n_jobs × bytes_per_voxel

      where ``bytes_per_voxel = n_gradients × 4 + output_bytes_per_voxel``.

    * **Parallelism upper bound** — at least ``n_jobs`` chunks are produced
      so all workers stay busy::

          N_par = n_vox // n_jobs   (only applied when n_vox is given)

    Final result: ``clamp(min(N_mem, N_par), lo=N_min)``.  Available RAM
    is detected via :func:`_available_ram` (stdlib only, no extra
    dependencies).

    Parameters
    ----------
    n_jobs : int
        Number of parallel Ray workers (concurrent in-flight tasks).
    n_gradients : int
        Number of diffusion gradients — sets the input chunk byte size.
    n_vox : int, optional
        Total number of voxels to fit.  Used to cap chunk size so that
        at least ``n_jobs`` chunks are produced.
    shared_obj_nbytes : int, optional
        Total bytes already placed in the Ray object store (e.g. the
        simulation library).  Deducted from the available RAM budget.
    ram_fraction : float, optional
        Fraction of available RAM to allocate to the Ray object store.
    min_efficiency : float, optional
        Target minimum ratio ``compute_time / total_task_time``.
    output_bytes_per_voxel : int, optional
        Estimated output bytes per voxel in the raw result dict returned
        by workers.  Defaults to 4 096 (≈ FORCE with ODF enabled).

    Returns
    -------
    chunk_size : int
    """
    n_min = int(_RAY_TASK_OVERHEAD_MS / (min_efficiency * _RAY_T_VOXEL_MS))

    n_jobs = determine_num_processes(n_jobs if n_jobs != 0 else 1)

    available = _available_ram()

    ray_budget = max(available * ram_fraction - shared_obj_nbytes, 0)
    bytes_per_voxel = n_gradients * 4 + output_bytes_per_voxel
    n_mem = int(ray_budget / n_jobs / bytes_per_voxel)

    n_par = (n_vox // n_jobs) if n_vox is not None else n_mem

    return max(n_min, min(n_mem, n_par))


@warning_for_keywords()
def paramap(
    func,
    in_list,
    *,
    out_shape=None,
    n_jobs=-1,
    engine="ray",
    backend=None,
    func_args=None,
    clean_spill=True,
    func_kwargs=None,
    shared_objects=None,
    inflight_cap=None,
    verbose=False,
    **kwargs,
):
    # FIXME: but several fitting functions return "extra" as well
    #        is this being handled properly, or not?
    #        perhaps the 'func' docs below are misleading?
    """
    Maps a function to a list of inputs in parallel.

    Parameters
    ----------
    func : callable
        The function to apply to each item in the array. Must have the form:
        ``func(arr, idx, *args, *kwargs)`` where arr is an ndarray and idx is
        an index into that array (a tuple). The Return of `func` needs to be
        one item (e.g. float, int) per input item.
    in_list : list
       A sequence of items each of which can be an input to ``func``.
    out_shape : tuple, optional
         The shape of the output array. If not specified, the output shape will
         be `(len(in_list),)`.
    n_jobs : int, optional
        The number of jobs to perform in parallel. Use -1 to use all but one
        cpu.
    engine : str, optional
        {"dask", "joblib", "ray", "serial"}
        The last one is useful for debugging -- runs the code without any
        parallelization.
    backend : str, optional
        The joblib or dask backend to use. For joblib the options are "loky",
        "threading", and "multiprocessing". For dask the options are
        "threading" and "multiprocessing".
    clean_spill : bool, optional
        If True, clean up the spill directory after the computation is done.
        Only applies to "ray" engine.
    func_args : list, optional
        Positional arguments to `func`.
    func_kwargs : dict or sequence, optional
        Keyword arguments to `func` or sequence of keyword arguments
        to `func`: one item for each item in the input list.
    shared_objects : dict, optional
        Dictionary of ``{name: array}`` to place in the Ray object store
        once via ``ray.put()``.  Each ``ObjectRef`` is then passed to
        workers as a direct top-level keyword argument prefixed with
        ``_sobj_`` (e.g. ``_sobj__index``).  Ray auto-resolves top-level
        ``ObjectRef`` arguments before invoking the worker — no explicit
        ``ray.get()`` is required on the worker side.  Only used with the
        ``"ray"`` engine; ignored for other engines.
    inflight_cap : int, optional
        Maximum number of pending Ray tasks before draining results.
        Limits memory pressure when processing many chunks.
        Only used with the ``"ray"`` engine; ignored for other engines.
        When None, all tasks are submitted at once.
    verbose : bool, optional
        Show a tqdm progress bar while processing chunks.
    kwargs : dict, optional
        Additional arguments to pass to either joblib.Parallel
        or dask.compute, or ray.remote depending on the engine used.

    Returns
    -------
    ndarray of identical shape to `arr`

    """
    func_args = func_args or []
    func_kwargs = func_kwargs or {}

    if isinstance(func_kwargs, Sequence):
        if len(func_kwargs) != len(in_list):
            raise ValueError(
                "The length of func_kwargs should be the same as the length of in_list"
            )
        func_kwargs_sequence = True
    else:
        func_kwargs_sequence = False

    if engine == "joblib":
        if not has_joblib:
            raise joblib()
        if backend is None:
            backend = "loky"
        pp = joblib.Parallel(n_jobs=n_jobs, backend=backend, **kwargs)
        dd = joblib.delayed(func)
        if func_kwargs_sequence:
            d_l = [dd(ii, *func_args, **fk) for ii, fk in zip(in_list, func_kwargs)]
        else:
            d_l = [dd(ii, *func_args, **func_kwargs) for ii in in_list]
        results = pp(tqdm(d_l, disable=not verbose))

    elif engine == "dask":
        if not has_dask:
            raise dask()
        if backend is None:
            backend = "threading"

        if n_jobs == -1:
            n_jobs = multiprocessing.cpu_count()
            n_jobs = n_jobs - 1

        def partial(func, *args, **keywords):
            def newfunc(in_arg):
                return func(in_arg, *args, **keywords)

            return newfunc

        delayed_func = dask.delayed(func)

        if func_kwargs_sequence:
            dd = [
                delayed_func(ii, *func_args, **fk)
                for ii, fk in zip(in_list, func_kwargs)
            ]
        else:
            pp = dask.delayed(partial(func, *func_args, **func_kwargs))
            dd = [pp(ii) for ii in in_list]

        if backend == "multiprocessing":
            results = dask.compute(*dd, scheduler="processes", workers=n_jobs, **kwargs)
        elif backend == "threading":
            results = dask.compute(*dd, scheduler="threads", workers=n_jobs, **kwargs)
        else:
            raise ValueError(f"{backend} is not a backend for dask")

    elif engine == "ray":
        if not has_ray:
            raise ray()

        if clean_spill:
            tmp_dir = tempfile.TemporaryDirectory()

            if not ray.is_initialized():
                ray.init(
                    _system_config={
                        "object_spilling_config": json.dumps(
                            {
                                "type": "filesystem",
                                "params": {"directory_path": tmp_dir.name},
                            }
                        )
                    }
                )

        shared_refs = None
        if shared_objects:
            shared_refs = {k: ray.put(v) for k, v in shared_objects.items()}

        func = ray.remote(func)

        def _build_kwargs(base_kw):
            if shared_refs is not None:
                sobj_kw = {f"_sobj_{k}": v for k, v in shared_refs.items()}
                return {**base_kw, **sobj_kw}
            return base_kw

        def _submit_one(item, kw):
            return func.remote(item, *func_args, **_build_kwargs(kw))

        n_chunks = len(in_list)
        if inflight_cap is not None and inflight_cap > 0:
            results = []
            pending = []
            items_kw = (
                zip(in_list, func_kwargs)
                if func_kwargs_sequence
                else ((ii, func_kwargs) for ii in in_list)
            )
            with tqdm(
                total=n_chunks, disable=not verbose, desc="Fitting (ray)"
            ) as pbar:
                for item, kw in items_kw:
                    pending.append(_submit_one(item, kw))
                    if len(pending) >= inflight_cap:
                        results.extend(ray.get(pending))
                        pbar.update(len(pending))
                        pending = []
                if pending:
                    results.extend(ray.get(pending))
                    pbar.update(len(pending))
        else:
            if func_kwargs_sequence:
                futures = [_submit_one(ii, fk) for ii, fk in zip(in_list, func_kwargs)]
            else:
                futures = [_submit_one(ii, func_kwargs) for ii in in_list]
            future_to_idx = {f: i for i, f in enumerate(futures)}
            results = [None] * n_chunks
            with tqdm(
                total=n_chunks, disable=not verbose, desc="Fitting (ray)"
            ) as pbar:
                remaining = list(futures)
                while remaining:
                    done, remaining = ray.wait(remaining, num_returns=1)
                    for f in done:
                        results[future_to_idx.pop(f)] = ray.get(f)
                        pbar.update(1)

        if clean_spill:
            shutil.rmtree(tmp_dir.name)

    elif engine == "serial":
        results = []
        with tqdm(
            total=len(in_list), disable=not verbose, desc="Fitting (serial)"
        ) as pbar:
            if func_kwargs_sequence:
                for in_element, fk in zip(in_list, func_kwargs):
                    results.append(func(in_element, *func_args, **fk))
                    pbar.update(1)
            else:
                for in_element in in_list:
                    results.append(func(in_element, *func_args, **func_kwargs))
                    pbar.update(1)
    else:
        raise ValueError("%s is not a valid engine" % engine)

    if out_shape is not None:
        return np.array(results).reshape(out_shape)
    else:
        return results



"""Testing reconstruction utilities."""

import numpy as np
from numpy.testing import assert_array_equal, assert_equal, assert_raises

from dipy.reconst.recspeed import adj_to_countarrs, argmax_from_countarrs


def test_adj_countarrs():
    adj = [[0, 1, 2], [2, 3], [4, 5, 6, 7]]
    counts, inds = adj_to_countarrs(adj)
    assert_array_equal(counts, [3, 2, 4])
    assert_equal(counts.dtype.type, np.uint32)
    assert_array_equal(inds, [0, 1, 2, 2, 3, 4, 5, 6, 7])
    assert_equal(inds.dtype.type, np.uint32)


def test_argmax_from_countarrs():
    # basic case
    vals = np.arange(10, dtype=float)
    vertinds = np.arange(10, dtype=np.uint32)
    adj_counts = np.ones((10,), dtype=np.uint32)
    adj_inds_raw = np.arange(10, dtype=np.uint32)[::-1]
    # when contiguous - OK
    adj_inds = adj_inds_raw.copy()
    argmax_from_countarrs(vals, vertinds, adj_counts, adj_inds)
    # yield assert_array_equal(inds, [5, 6, 7, 8, 9])

    # test for errors - first - not contiguous
    with assert_raises(ValueError):
        argmax_from_countarrs(vals, vertinds, adj_counts, adj_inds_raw)

    # too few vertices
    with assert_raises(ValueError):
        argmax_from_countarrs(vals, vertinds[:-1], adj_counts, adj_inds)

    # adj_inds too short
    with assert_raises(IndexError):
        argmax_from_countarrs(vals, vertinds, adj_counts, adj_inds[:-1])

    # vals too short
    with assert_raises(IndexError):
        argmax_from_countarrs(vals[:-1], vertinds, adj_counts, adj_inds)



# cython: boundscheck=False, wraparound=False, cdivision=True
"""
Cython backend for bias field correction.

Provides accelerated B-spline design matrix construction, Gram matrix
accumulation, and Tukey biweight weight computation.
"""

cimport numpy as cnp
import numpy as np


cdef void _bspline_1d_values(
    double t,
    long n_ctrl,
    double* bvals,
    long* start_idx,
) noexcept nogil:
    """Evaluate cubic B-spline basis at a single parameter value.

    Parameters
    ----------
    t : double
        Parameter value in [0, n_ctrl - 1).  Clamped to this range if
        out of bounds.
    n_ctrl : long
        Number of control points along this axis.
    bvals : double*
        Output array of length 4.  Filled with the four non-zero cubic
        B-spline basis values at ``t``.
    start_idx : long*
        Output scalar.  Set to the control-point index corresponding to
        ``bvals[0]``; the active indices are ``start_idx[0] + 0..3``.
        Boundary control points (index < 0 or >= n_ctrl) must be
        discarded by the caller.

    Returns
    -------
    None
        Results are written into ``bvals`` and ``start_idx`` in place.
    """
    cdef:
        long k
        double u, u2, u3

    if t < 0.0:
        t = 0.0
    if t >= n_ctrl - 1:
        t = n_ctrl - 1 - 1e-10

    k = <long>t
    if k >= n_ctrl - 1:
        k = n_ctrl - 2

    u = t - k
    u2 = u * u
    u3 = u2 * u

    bvals[0] = (1.0 - u) * (1.0 - u) * (1.0 - u) / 6.0
    bvals[1] = (3.0 * u3 - 6.0 * u2 + 4.0) / 6.0
    bvals[2] = (-3.0 * u3 + 3.0 * u2 + 3.0 * u + 1.0) / 6.0
    bvals[3] = u3 / 6.0

    start_idx[0] = k - 1


def masked_voxel_coords(
    unsigned char[:, :, ::1] mask,
    cnp.int64_t[:, ::1] out,
):
    """Extract (i, j, k) voxel coordinates of all nonzero mask entries.

    Parameters
    ----------
    mask : ndarray
        3D uint8 mask, C-contiguous, shape (S, R, C).
    out : ndarray
        Pre-allocated output array, shape (N_max, 3), dtype int64.
        Filled with (i, j, k) coordinates of nonzero voxels in
        row-major order.

    Returns
    -------
    count : int
        Number of nonzero voxels found (actual rows written to ``out``).
    """
    cdef:
        long S = mask.shape[0]
        long R = mask.shape[1]
        long C = mask.shape[2]
        long i, j, k, count

    count = 0
    with nogil:
        for i in range(S):
            for j in range(R):
                for k in range(C):
                    if mask[i, j, k] != 0:
                        out[count, 0] = i
                        out[count, 1] = j
                        out[count, 2] = k
                        count += 1
    return count


def evaluate_bspline_rows(
    double[:, ::1] grid_coords,
    cnp.int64_t[::1] n_ctrl,
    cnp.int64_t[::1] row_ptr,
    cnp.int64_t[::1] col_idx,
    double[::1] values,
):
    """Evaluate cubic B-spline basis for N masked voxels, filling CSR arrays.

    Parameters
    ----------
    grid_coords : ndarray
        Voxel positions in control grid space, shape (N, 3), C-contiguous,
        dtype float64.  Axis order is (z, y, x).
    n_ctrl : ndarray
        Control grid dimensions (ns, nr, nc), shape (3,), dtype int64.
    row_ptr : ndarray
        Output CSR row pointers, shape (N+1,), dtype int64.
        Pre-allocated; written in-place.
    col_idx : ndarray
        Output CSR column indices, shape (N*64,), dtype int64.
        Pre-allocated; written in-place.
    values : ndarray
        Output CSR non-zero values, shape (N*64,), dtype float64.
        Pre-allocated; written in-place.

    Returns
    -------
    nnz : int
        Actual number of non-zero entries written to ``col_idx`` and
        ``values``.  Always <= N * 64.
    """
    cdef:
        long N = grid_coords.shape[0]
        long ns = n_ctrl[0]
        long nr = n_ctrl[1]
        long nc = n_ctrl[2]
        long row, nnz, iz, iy, ix, cz, cy, cx, col
        double bz_vals[4]
        double by_vals[4]
        double bx_vals[4]
        long start_z, start_y, start_x
        double vz, vy, vx

    nnz = 0
    row_ptr[0] = 0

    with nogil:
        for row in range(N):
            _bspline_1d_values(grid_coords[row, 0], ns, bz_vals, &start_z)
            _bspline_1d_values(grid_coords[row, 1], nr, by_vals, &start_y)
            _bspline_1d_values(grid_coords[row, 2], nc, bx_vals, &start_x)

            for iz in range(4):
                cz = start_z + iz
                if cz < 0 or cz >= ns:
                    continue
                vz = bz_vals[iz]

                for iy in range(4):
                    cy = start_y + iy
                    if cy < 0 or cy >= nr:
                        continue
                    vy = by_vals[iy]

                    for ix in range(4):
                        cx = start_x + ix
                        if cx < 0 or cx >= nc:
                            continue
                        vx = bx_vals[ix]

                        col = cz * nr * nc + cy * nc + cx
                        col_idx[nnz] = col
                        values[nnz] = vz * vy * vx
                        nnz += 1

            row_ptr[row + 1] = nnz

    return nnz


def gram_matrix_csr(
    double[::1] data,
    int[::1] indices,
    int[::1] indptr,
    double[::1] weights,
    double[::1] y,
    double[:, ::1] A,
    double[::1] b_vec,
):
    """Accumulate weighted Gram matrix and right-hand side from a CSR matrix.

    Computes in-place::

        A[j, k]   += sum_i  w_i * X[i,j] * X[i,k]
        b_vec[k]  += sum_i  w_i * y_i    * X[i,k]

    Rows with ``weights[i] == 0`` are skipped.

    Parameters
    ----------
    data : ndarray
        CSR non-zero values, shape (nnz,), dtype float64.
    indices : ndarray
        CSR column indices, shape (nnz,), dtype int32.
    indptr : ndarray
        CSR row pointers, shape (N+1,), dtype int32.
    weights : ndarray
        Per-row regression weights, shape (N,), dtype float64.
    y : ndarray
        Target values, shape (N,), dtype float64.
    A : ndarray
        K x K Gram matrix, pre-allocated zeros, shape (K, K), dtype
        float64.  Updated in-place.
    b_vec : ndarray
        Right-hand side vector, pre-allocated zeros, shape (K,), dtype
        float64.  Updated in-place.

    Returns
    -------
    None
        ``A`` and ``b_vec`` are modified in-place; nothing is returned.
    """
    cdef:
        long N = weights.shape[0]
        long i
        int p, q, jp, jq, start, end, nnz_i
        double w_i, wy_i, dp, dq

    with nogil:
        for i in range(N):
            w_i = weights[i]
            if w_i == 0.0:
                continue
            wy_i = w_i * y[i]
            start = indptr[i]
            end = indptr[i + 1]
            nnz_i = end - start

            for p in range(nnz_i):
                jp = indices[start + p]
                dp = data[start + p]
                b_vec[jp] += wy_i * dp

                for q in range(nnz_i):
                    jq = indices[start + q]
                    dq = data[start + q]
                    A[jp, jq] += w_i * dp * dq


def compute_tukey_weights(
    double[::1] residuals,
    double[::1] weights,
    *,
    double c=4.685,
):
    """Compute Tukey biweight weights and multiply into ``weights`` in-place.

    Estimates the residual scale via the Median Absolute Deviation (MAD),
    then applies the Tukey bisquare function::

        weights[i] *= (1 - (r_i / (c * MAD))^2)^2   if |r_i| < c * MAD
        weights[i]  = 0                               otherwise

    where ``MAD = median(|residuals|) / 0.6745``.

    Parameters
    ----------
    residuals : ndarray
        Regression residuals, C-contiguous float64, shape (N,).
    weights : ndarray
        Existing per-voxel weights to update in-place, C-contiguous
        float64, shape (N,).
    c : float
        Tukey breakdown constant.  The default (4.685) gives 95 %
        efficiency under Gaussian noise.

    Returns
    -------
    None
        ``weights`` is modified in-place; nothing is returned.
    """
    cdef:
        long N = residuals.shape[0]
        long i
        double mad, scale, u, w_val

    if N == 0:
        return

    abs_res_np = np.abs(np.asarray(residuals))
    mad = float(np.median(abs_res_np)) / 0.6745

    if mad < 1e-15:
        return

    scale = c * mad

    for i in range(N):
        u = residuals[i] / scale
        if u < -1.0 or u > 1.0:
            weights[i] = 0.0
        else:
            w_val = 1.0 - u * u
            weights[i] = weights[i] * w_val * w_val



"""

Note
----

This file is a pytorch adapted version from the
sources of the SynthSeg project - https://github.com/BBillot/SynthSeg.
All weights and model artchitecture are original from the SynthSeg project.
It remains licensed as the rest of SynthSeg
(Apache 2.0 license as of January 2026).

# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE.txt file distributed along with the SynthSeg package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""

import numpy as np
from scipy.ndimage import gaussian_filter

from dipy.data import get_fnames
from dipy.nn.utils import (
    normalize,
    recover_img,
    set_logger_level,
    transform_img,
)
from dipy.segment.utils import remove_holes_and_islands
from dipy.testing.decorators import doctest_skip_parser
from dipy.utils.logging import logger
from dipy.utils.optpkg import optional_package

torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
if have_torch:
    from torch.nn import (
        BatchNorm3d,
        Conv3d,
        MaxPool3d,
        Module,
        ModuleList,
        Softmax,
        Upsample,
    )
    import torch.nn.functional as F
else:

    class Module:
        pass


class Conv3dELU(Module):
    """
    Mimics TensorFlow Conv3D + ELU fused behavior.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        padding=0,
        stride=1,
        dilation=1,
        bias=True,
    ):
        super().__init__()
        self.conv = Conv3d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            padding=padding,
            stride=stride,
            dilation=dilation,
            bias=bias,
        )

    def forward(self, x):
        """Forward pass of the Conv3dELU layer.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor after convolution and ELU activation.
        """
        x = self.conv(x)
        x = F.elu(x)
        return x


class Block(Module):
    """Building block for the SynthSeg model.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    kernel_size : int
        Size of the convolutional kernel.
    padding : int
        Padding of the convolution.
    n_layers : int
        Number of convolutional layers in the block.
    layer_type : str
        Type of the block: 'down', 'down2', or 'up'.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        padding,
        n_layers,
        layer_type,
    ):
        super(Block, self).__init__()
        self.n_layers = n_layers
        self.layer_list = ModuleList()
        self.layer_type = layer_type

        cur_channel = in_channels
        if self.layer_type == "up":
            self.layer_list.append(Upsample(scale_factor=2))
            cur_channel = int(cur_channel * 1.5)

        for _ in range(n_layers):
            self.layer_list.append(
                Conv3dELU(
                    cur_channel,
                    out_channels,
                    kernel_size,
                    padding=padding,
                )
            )
            cur_channel = out_channels

        self.layer_list.append(BatchNorm3d(out_channels, eps=1e-3, momentum=0.01))
        if self.layer_type == "down":
            self.layer_list.append(MaxPool3d(2, stride=2))

    def forward(self, input, passed):
        """Forward pass of the Block.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor.
        passed : torch.Tensor
            Tensor from skipped connection.

        Returns
        -------
        x : torch.Tensor
            Output of the convolutional layers.
        skip : torch.Tensor, optional
            Output of the convolutional layers before downsampling.
            Only returned if layer_type is 'down'.
        """
        x = input
        for l_idx, layer in enumerate(self.layer_list):
            x = layer(x)
            if l_idx == 0 and self.layer_type == "up":
                x = torch.cat([passed, x], dim=1)
            if l_idx == self.n_layers - 1 and self.layer_type == "down":
                skip = x

        if self.layer_type == "down":
            return x, skip
        return x


class Model(Module):
    """The Model class for the SynthSeg architecture.

    Parameters
    ----------
    model_scale : int, optional
        The scale of the model.
    n_levels : int, optional
        Number of levels in the U-Net.
    output_channels : int, optional
        Number of output channels.
    """

    def __init__(self, *, model_scale=24, n_levels=5, output_channels=33):
        super(Model, self).__init__()

        self.block_list = ModuleList()
        self.channels = [model_scale * (2**i) for i in range(n_levels)]
        self.channels_rev = [
            model_scale * (2 ** (n_levels - i - 1)) for i in range(n_levels)
        ]
        self.channels.insert(0, 1)
        self.channels_rev.append(output_channels)

        self.model_scale = model_scale
        self.n_levels = n_levels

        # Block structure
        for level in range(n_levels - 1):
            block = Block(
                self.channels[level],
                self.channels[level + 1],
                kernel_size=3,
                padding=1,
                n_layers=2,
                layer_type="down",
            )
            self.block_list.append(block)

        level = n_levels - 1
        block = Block(
            self.channels[level],
            self.channels[level + 1],
            kernel_size=3,
            padding=1,
            n_layers=2,
            layer_type="down2",
        )
        self.block_list.append(block)

        for level in range(n_levels - 1):
            block = Block(
                self.channels_rev[level],
                self.channels_rev[level + 1],
                kernel_size=3,
                padding=1,
                n_layers=2,
                layer_type="up",
            )
            self.block_list.append(block)

        self.conv_pred = Conv3d(model_scale, 33, 1, padding=0)
        self.softmax = Softmax(dim=1)

    def forward(self, inputs):
        """Forward pass of the SynthSeg model.

        Parameters
        ----------
        inputs : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Predicted brain segmentation probability map.
        """
        skip_list = []
        x = inputs
        for block in self.block_list[: self.n_levels - 1]:
            x, skip = block(x, None)
            skip_list.append(skip)
        x = self.block_list[self.n_levels - 1](x, None)
        for idx, block in enumerate(self.block_list[self.n_levels :]):
            passed = skip_list[-(idx + 1)]
            x = block(x, passed)
        x = self.conv_pred(x)
        output = self.softmax(x)

        return output


class SynthSeg:
    """This class is intended for the SynthSeg model.

    The SynthSeg model :footcite:p:`Billot2023` is a deep learning neural network for
    brain segmentation. It uses a U-net architecture and was trained on synthetic
    images generated from label maps. The model is robust to variations in
    contrast and resolution, making it suitable for segmenting a wide range of
    brain scans.

    Note that we are not saving any PVE maps here, only the hard segmentation
    due to the size of the output probability maps.

    References
    ----------
    .. footbibliography::
    """

    @doctest_skip_parser
    def __init__(self, *, verbose=False, use_cuda=False):
        """Model initialization

        Parameters
        ----------
        verbose : bool, optional
            Whether to show information about the processing.
        use_cuda : bool, optional
            Whether to use GPU for processing.
            If False or no CUDA is detected, CPU will be used.
        """
        if not have_torch:
            raise torch()

        log_level = "INFO" if verbose else "CRITICAL"
        set_logger_level(log_level, logger)

        self.model = self.init_model()
        self.model.eval()
        if use_cuda:
            if torch.cuda.is_available():
                self.device = torch.device("cuda")
            else:
                logger.warning("CUDA requested but not found, switching to CPU")
                self.device = torch.device("cpu")
        else:
            self.device = torch.device("cpu")
        self.model = self.model.to(self.device)
        self.fetch_default_weights()
        self.labels_segmentation = np.array(
            [
                0,
                2,
                3,
                4,
                5,
                7,
                8,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                24,
                26,
                28,
                41,
                42,
                43,
                44,
                46,
                47,
                49,
                50,
                51,
                52,
                53,
                54,
                58,
                60,
            ]
        )
        self.topological_classes = np.array(
            [
                0,
                4,
                4,
                4,
                4,
                5,
                5,
                6,
                7,
                8,
                9,
                1,
                2,
                3,
                10,
                11,
                0,
                12,
                13,
                14,
                14,
                14,
                14,
                15,
                15,
                16,
                17,
                18,
                19,
                20,
                21,
                22,
                23,
            ]
        )
        self.flip_indices = np.array(
            [
                0,
                19,
                20,
                21,
                22,
                23,
                24,
                25,
                26,
                27,
                28,
                11,
                12,
                13,
                29,
                30,
                16,
                31,
                32,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                14,
                15,
                17,
                18,
            ]
        )
        self.label_dict = {
            0: "background",
            2: "left cerebral white matter",
            3: "left cerebral cortex",
            4: "left lateral ventricle",
            5: "left inferior lateral ventricle",
            7: "left cerebellum white matter",
            8: "left cerebellum cortex",
            10: "left thalamus",
            11: "left caudate",
            12: "left putamen",
            13: "left pallidum",
            14: "3rd ventricle",
            15: "4th ventricle",
            16: "brain-stem",
            17: "left hippocampus",
            18: "left amygdala",
            24: "CSF",
            26: "left accumbens area",
            28: "left ventral DC",
            41: "right cerebral white matter",
            42: "right cerebral cortex",
            43: "right lateral ventricle",
            44: "right inferior lateral ventricle",
            46: "right cerebellum white matter",
            47: "right cerebellum cortex",
            49: "right thalamus",
            50: "right caudate",
            51: "right putamen",
            52: "right pallidum",
            53: "right hippocampus",
            54: "right amygdala",
            58: "right accumbens area",
            60: "right ventral DC",
        }

    def init_model(self, model_scale=24, n_levels=5, output_channels=33):
        """Initialize the SynthSeg model.

        Parameters
        ----------
        model_scale : int, optional
            The scale of the model.
        n_levels : int, optional
            Number of levels in the U-Net.
        output_channels : int, optional
            Number of output channels.

        Returns
        -------
        Model
            Initialized SynthSeg model.
        """
        return Model(
            model_scale=model_scale, n_levels=n_levels, output_channels=output_channels
        )

    def fetch_default_weights(self):
        """Load the model pre-training weights to use for the fitting.

        While the user can load different weights, the function
        is mainly intended for the class function 'predict'.
        """
        fetch_model_weights_path = get_fnames(name="synthseg_torch_weights")
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        """Load the custom pre-training weights to use for the fitting.

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights (pth, saved by Pytorch)
        """
        try:
            self.model.load_state_dict(
                torch.load(
                    weights_path,
                    weights_only=True,
                    map_location=self.device,
                )
            )
            self.model.eval()
        except ValueError as e:
            raise ValueError(
                "Expected input for the provided model weights \
                             do not match the declared model"
            ) from e

    def _flip_img_indices(self, img):
        """Flip the label indices for a flipped image.

        Parameters
        ----------
        img : np.ndarray
            The image with label indices.

        Returns
        -------
        np.ndarray
            The image with flipped label indices.
        """
        new_img = np.zeros_like(img)
        for l_idx, flip_lbl in enumerate(self.flip_indices):
            if l_idx != 0:
                new_img = np.where(img == l_idx, flip_lbl, new_img)
        return new_img

    def _prepare_img(self, img):
        """Prepare the image for model input.

        Parameters
        ----------
        img : np.ndarray
            Input image.

        Returns
        -------
        torch.Tensor
            Image tensor ready for the model.
        """
        img = np.expand_dims(img, 1)  # add channel dimension
        img_tensor = torch.tensor(img, dtype=torch.float32).to(self.device)
        return img_tensor

    def __predict(self, x_test):
        """Internal prediction function

        Parameters
        ----------
        x_test : list of np.ndarray
            Image should match the required shape of the model.

        Returns
        -------
        np.ndarray (batch, ...)
            Predicted brain labels
        """
        return self.model(x_test).detach().numpy()

    def predict(
        self,
        T1,
        affine,
        *,
        batch_size=None,
        return_prob=False,
    ):
        """Wrapper function to facilitate prediction of larger dataset.

        Parameters
        ----------
        T1 : np.ndarray or list of np.ndarray
            For a single image, input should be a 3D array.
            If multiple images, it should be a a list or tuple.
        affine : np.ndarray (4, 4) or (batch, 4, 4)
            or list of np.ndarrays with len of batch
            Affine matrix for the T1 image. Should have
            batch dimension if T1 has one.
        batch_size : int, optional
            Number of images per prediction pass. Only available if data
            is provided with a batch dimension.
            Consider lowering it if you get an out of memory error.
            Increase it if you want it to be faster and have a lot of data.
            If None, batch_size will be set to 1 if the provided image
            has a batch dimension.
        return_prob : bool, optional
            Whether to return the probability map instead of a
            label map. Useful for testing.

        Returns
        -------
        pred_output : np.ndarray (...) or (batch, ...)
            Predicted brain labels. If return_prob is True, it will be a
            probability map instead of a label map.

        label_dict : dict
            Dictionary mapping label indices to anatomical structure names.
            Only if return_prob is False.

        mask : np.ndarray (...) or (batch, ...)
            Predicted brain mask.
            Only if return_prob is False.
        """
        affine = np.array(affine)

        if isinstance(T1, (list, tuple)):
            dim = 4
            T1 = np.array(T1)
        elif len(T1.shape) == 3:
            dim = 3
            if batch_size is not None:
                logger.warning(
                    "Batch size specified, but not used",
                    "due to the input not having \
                                a batch dimension",
                )

            T1 = np.expand_dims(T1, 0)
            affine = np.expand_dims(affine, 0)
        else:
            raise ValueError(
                "Input data should be a np.ndarray of dimension 3 or a list/tuple of it"
            )
        if batch_size is None:
            batch_size = 1

        input_data = np.zeros((len(T1), 192, 192, 192))
        params_list = []

        # Normalize the data.
        ori_shape = T1.shape[1:]
        for i, T1_img in enumerate(T1):
            t_img, params = transform_img(
                T1_img,
                affine[i],
                target_voxsize=(1.0, 1.0, 1.0),
                final_size=(192, 192, 192),
                order=3,
            )
            min_v, max_v = np.percentile(t_img, (0.5, 99.5))
            t_img = normalize(t_img, min_v=min_v, max_v=max_v, new_min=0, new_max=1)
            input_data[i] = t_img
            params_list.append(params)
        # Prediction stage
        prediction = np.zeros((len(T1), 192, 192, 192, 33), dtype=np.float32)
        for batch_idx in range(batch_size, len(T1) + 1, batch_size):
            batch = input_data[batch_idx - batch_size : batch_idx]
            batch_start = batch_idx - batch_size
            temp_input = self._prepare_img(batch)
            temp_pred = self.__predict(temp_input)
            temp_pred = gaussian_filter(temp_pred, (0, 0.5, 0.5, 0.5, 0))
            prediction[batch_start:batch_idx] = np.moveaxis(temp_pred, 1, -1)
            temp_input = self._prepare_img(np.flip(batch, axis=1).copy())
            temp_pred = self.__predict(temp_input)
            temp_pred = gaussian_filter(temp_pred, (0, 0.5, 0.5, 0.5, 0))
            temp_pred = np.flip(temp_pred, axis=2)
            temp_pred = np.stack(
                [
                    temp_pred[:, self.flip_indices[i]]
                    for i in range(len(self.flip_indices))
                ],
                axis=1,
            )
            prediction[batch_start:batch_idx] += np.moveaxis(temp_pred, 1, -1)
            prediction[batch_start:batch_idx] /= 2
        remainder = np.mod(len(T1), batch_size)
        if remainder != 0:
            batch = input_data[-remainder:]
            temp_input = self._prepare_img(batch)
            temp_pred = self.__predict(temp_input)
            temp_pred = gaussian_filter(temp_pred, (0, 0.5, 0.5, 0.5, 0))
            prediction[-remainder:] = np.moveaxis(temp_pred, 1, -1)
            temp_input = self._prepare_img(np.flip(batch, axis=1).copy())
            temp_pred = self.__predict(temp_input)
            temp_pred = gaussian_filter(temp_pred, (0, 0.5, 0.5, 0.5, 0))
            temp_pred = np.flip(temp_pred, axis=1)
            temp_pred = np.stack(
                [
                    temp_pred[:, self.flip_indices[i]]
                    for i in range(len(self.flip_indices))
                ],
                axis=1,
            )
            prediction[-remainder:] += np.moveaxis(temp_pred, 1, -1)
            prediction[-remainder:] /= 2

        if return_prob:
            labels = np.zeros((len(T1),) + (192, 192, 192, 33)).astype(np.float32)
        else:
            labels = np.zeros((len(T1),) + ori_shape).astype(np.int32)
            masks = np.zeros((len(T1),) + ori_shape)
        for i in range(len(T1)):
            output = prediction[i]

            tmp_post_patch_seg = output[..., 1:]
            post_patch_seg_mask = np.sum(tmp_post_patch_seg, axis=-1) > 0.25
            post_patch_seg_mask = remove_holes_and_islands(
                post_patch_seg_mask, remove_holes=False
            ).astype(bool)
            output[..., 1:] = output[..., 1:] * np.stack(
                [post_patch_seg_mask] * 32, axis=-1
            )

            post_patch_seg_mask = output > 0.25
            for topology_class in np.unique(self.topological_classes)[1:]:
                tmp_topology_indices = np.where(
                    self.topological_classes == topology_class
                )[0]
                tmp_mask = np.any(
                    post_patch_seg_mask[..., tmp_topology_indices], axis=-1
                )
                tmp_mask = remove_holes_and_islands(
                    tmp_mask, remove_holes=False
                ).astype(bool)
                for idx in tmp_topology_indices:
                    output[..., idx] *= tmp_mask

            output /= np.sum(output, axis=-1)[..., np.newaxis]
            if return_prob:
                labels[i] = output.astype("float32")
                continue
            else:
                temp = self.labels_segmentation[
                    output.argmax(-1).astype("int32")
                ].astype("int32")
                temp = recover_img(temp, params_list[i], order=0)
                labels[i] = np.round(temp).astype(np.int32)
            masks[i] = (labels[i] > 0).astype(np.int32)

        if dim == 3:
            labels = labels[0]
            if not return_prob:
                masks = masks[0]

        if return_prob:
            return labels
        return labels, self.label_dict, masks



"""Bias field correction for diffusion MRI data.

Provides classical regression-based bias field correction via Legendre
polynomial regression and cubic B-spline regression.

The bias field is estimated exclusively from the mean b0 volume in the
log domain and applied uniformly to all DWI volumes.
"""

import numpy as np
from scipy import linalg as scipy_linalg, ndimage, sparse

try:
    from dipy.denoise._bias_correction import (
        compute_tukey_weights,
        evaluate_bspline_rows,
        gram_matrix_csr,
        masked_voxel_coords,
    )

    _HAVE_CYTHON = True
except ImportError:
    _HAVE_CYTHON = False

from dipy.core.gradients import extract_b0
from dipy.segment.mask import applymask, median_otsu
from dipy.utils.logging import logger

try:
    from dipy.align.vector_fields import gradient as _vf_gradient

    _HAVE_VF_GRADIENT = True
except ImportError:
    _HAVE_VF_GRADIENT = False


def _get_mean_b0(data, gtab):
    """Return mean b0 volume as float64.

    Parameters
    ----------
    data : ndarray
        4D DWI data (X, Y, Z, N).
    gtab : GradientTable
        Gradient table with b0s_mask attribute.

    Returns
    -------
    mean_b0 : ndarray
        3D mean b0 volume, dtype float64.
    """
    return extract_b0(data, gtab.b0s_mask, strategy="mean").astype(np.float64)


def _get_mask(mean_b0, mask):
    """Return binary brain mask, computing via median_otsu if not provided.

    Parameters
    ----------
    mean_b0 : ndarray
        3D mean b0 volume.
    mask : ndarray or None
        Existing 3D binary mask, or None to auto-compute.

    Returns
    -------
    mask : ndarray
        3D boolean brain mask.
    """
    if mask is None:
        _, mask = median_otsu(mean_b0, median_radius=4, numpass=4)
    else:
        mask = np.asarray(mask, dtype=bool)
    return mask


def _gradient_weights(*, log_b0, alpha=1.0):
    """Compute gradient-based edge suppression weight map.

    Parameters
    ----------
    log_b0 : ndarray
        3D log-domain b0 image, shape (S, R, C).
    alpha : float, optional
        Edge suppression strength.

    Returns
    -------
    weights : ndarray
        Float64 weight map, same shape as log_b0.
    """
    img = np.ascontiguousarray(log_b0, dtype=np.float64)
    if _HAVE_VF_GRADIENT:
        shape = np.array(img.shape, dtype=np.int32)
        eye4 = np.eye(4, dtype=np.float64)
        spacing = np.ones(3, dtype=np.float64)
        grad_out, _ = _vf_gradient(img, eye4, spacing, shape, eye4)
        grad_mag = np.sqrt(np.sum(grad_out**2, axis=-1))
    else:
        gx = ndimage.sobel(img, axis=0)
        gy = ndimage.sobel(img, axis=1)
        gz = ndimage.sobel(img, axis=2)
        grad_mag = np.sqrt(gx**2 + gy**2 + gz**2)
    return np.exp(-alpha * grad_mag)


def _normalize_coords(*, shape, coords):
    """Normalize voxel coordinates to [-1, 1] along each axis.

    Parameters
    ----------
    shape : tuple of int
        Volume shape (S, R, C).
    coords : ndarray
        Integer coordinates, shape (N, 3).

    Returns
    -------
    coords_norm : ndarray
        Normalized float64 coordinates, shape (N, 3).
    """
    coords_norm = coords.astype(np.float64)
    for d, n in enumerate(shape):
        if n > 1:
            coords_norm[:, d] = 2.0 * coords_norm[:, d] / (n - 1) - 1.0
        else:
            coords_norm[:, d] = 0.0
    return coords_norm


def _legendre_basis(*, coords_flat, order):
    """Build Legendre polynomial design matrix.

    Parameters
    ----------
    coords_flat : ndarray
        Normalized coordinates in [-1, 1], shape (N, 3).
    order : int
        Maximum total polynomial degree (terms where i+j+k <= order).

    Returns
    -------
    X : ndarray
        Design matrix, shape (N, K) where K is the number of terms.
    """
    from numpy.polynomial.legendre import legval

    terms = [
        (i, j, k)
        for i in range(order + 1)
        for j in range(order + 1 - i)
        for k in range(order + 1 - i - j)
    ]
    N = coords_flat.shape[0]
    K = len(terms)
    X = np.zeros((N, K), dtype=np.float64)

    for col, (i, j, k) in enumerate(terms):
        ei = np.zeros(i + 1)
        ei[i] = 1.0
        ej = np.zeros(j + 1)
        ej[j] = 1.0
        ek = np.zeros(k + 1)
        ek[k] = 1.0
        X[:, col] = (
            legval(coords_flat[:, 0], ei)
            * legval(coords_flat[:, 1], ej)
            * legval(coords_flat[:, 2], ek)
        )
    return X


def _weighted_ridge_solve(*, X, y, weights, lambda_reg):
    """Solve weighted ridge regression min ||W^(1/2)(y - Xβ)||² + λ||β||².

    Parameters
    ----------
    X : ndarray
        Design matrix, shape (N, K).
    y : ndarray
        Target values, shape (N,).
    weights : ndarray
        Non-negative regression weights, shape (N,).
    lambda_reg : float
        Ridge regularization strength.

    Returns
    -------
    beta : ndarray
        Coefficient vector, shape (K,).
    """
    K = X.shape[1]
    WX = weights[:, None] * X
    A = X.T @ WX + lambda_reg * np.eye(K)
    b = X.T @ (weights * y)
    try:
        beta = np.linalg.solve(A, b)
    except np.linalg.LinAlgError:
        beta, _, _, _ = np.linalg.lstsq(A, b, rcond=None)
    return beta


def _tukey_weights_py(*, residuals, c):
    """Compute Tukey biweight weights (pure Python/NumPy).

    Parameters
    ----------
    residuals : ndarray
        Regression residuals, shape (N,).
    c : float
        Tukey breakdown constant.

    Returns
    -------
    weights : ndarray
        Tukey biweight weights in [0, 1], shape (N,).
    """
    mad = np.median(np.abs(residuals)) / 0.6745
    if mad < 1e-15:
        return np.ones(len(residuals), dtype=np.float64)
    u = residuals / (c * mad)
    w = np.where(np.abs(u) < 1.0, (1.0 - u**2) ** 2, 0.0)
    return w.astype(np.float64)


def _tukey_weights(*, residuals, c=4.685):
    """Compute Tukey biweight weights, using Cython backend if available.

    Parameters
    ----------
    residuals : ndarray
        Regression residuals, shape (N,).
    c : float, optional
        Tukey breakdown constant.

    Returns
    -------
    weights : ndarray
        Tukey biweight weights in [0, 1], shape (N,).
    """
    if _HAVE_CYTHON:
        w = np.ones(len(residuals), dtype=np.float64)
        compute_tukey_weights(np.ascontiguousarray(residuals, dtype=np.float64), w, c=c)
        return w
    return _tukey_weights_py(residuals=residuals, c=c)


def _polynomial_pyramid_fit(
    *,
    log_b0,
    mask,
    order,
    pyramid_levels,
    n_iter,
    lambda_reg,
    robust,
    gradient_weighting,
    sigma_factor=0.2,
):
    """Coarse-to-fine polynomial bias field regression.

    Parameters
    ----------
    log_b0 : ndarray
        3D log-domain b0 image, shape (S, R, C).
    mask : ndarray
        3D boolean brain mask.
    order : int
        Maximum Legendre polynomial order.
    pyramid_levels : tuple of int
        Downsampling factors, ordered coarse-first (e.g. (4, 2, 1)).
    n_iter : int
        Reweighting iterations per pyramid level.
    lambda_reg : float
        Ridge regularization strength.
    robust : bool
        Apply Tukey biweight robust reweighting.
    gradient_weighting : bool
        Apply gradient-based edge suppression weights.
    sigma_factor : float, optional
        Sigma = factor * sigma_factor for Gaussian smoothing.

    Returns
    -------
    log_bias : ndarray
        Estimated log-domain bias field, same shape as log_b0.
    """
    full_shape = log_b0.shape

    if gradient_weighting:
        grad_w_full = _gradient_weights(log_b0=log_b0)

    ii, jj, kk = np.meshgrid(
        np.arange(full_shape[0]),
        np.arange(full_shape[1]),
        np.arange(full_shape[2]),
        indexing="ij",
    )
    full_vox_coords = np.column_stack([ii.ravel(), jj.ravel(), kk.ravel()])
    full_coords_norm = _normalize_coords(shape=full_shape, coords=full_vox_coords)
    X_full = _legendre_basis(coords_flat=full_coords_norm, order=order)

    # Center log_b0 so the polynomial fits only spatial variation, not the
    # DC offset (overall intensity level).
    log_b0_dc = log_b0[mask].mean()
    residual = log_b0 - log_b0_dc

    log_bias = np.zeros(full_shape, dtype=np.float64)

    for factor in pyramid_levels:
        if factor == 1:
            level_residual = residual
            level_mask = mask
        else:
            sigma = factor * sigma_factor
            smoothed = ndimage.gaussian_filter(residual, sigma=sigma)
            level_residual = ndimage.zoom(smoothed, zoom=1.0 / factor, order=1)
            level_mask = (
                ndimage.zoom(mask.astype(np.float64), zoom=1.0 / factor, order=0) > 0.5
            )

        level_shape = level_residual.shape
        mask_flat = level_mask.ravel()
        n_masked = mask_flat.sum()

        # Need at least as many data points as parameters
        n_params = sum(
            1
            for i in range(order + 1)
            for j in range(order + 1 - i)
            for _ in range(order + 1 - i - j)
        )
        if n_masked < n_params:
            continue

        y = level_residual.ravel()[mask_flat]

        ii_l, jj_l, kk_l = np.meshgrid(
            np.arange(level_shape[0]),
            np.arange(level_shape[1]),
            np.arange(level_shape[2]),
            indexing="ij",
        )
        level_coords = np.column_stack(
            [
                ii_l.ravel()[mask_flat],
                jj_l.ravel()[mask_flat],
                kk_l.ravel()[mask_flat],
            ]
        )
        coords_norm = _normalize_coords(shape=level_shape, coords=level_coords)
        X = _legendre_basis(coords_flat=coords_norm, order=order)

        w = np.ones(n_masked, dtype=np.float64)
        if gradient_weighting:
            if factor == 1:
                gw = grad_w_full.ravel()[mask_flat]
            else:
                gw_down = ndimage.zoom(grad_w_full, zoom=1.0 / factor, order=1)
                gw = gw_down.ravel()[mask_flat]
            w = w * gw

        beta = None
        for _ in range(n_iter):
            beta = _weighted_ridge_solve(X=X, y=y, weights=w, lambda_reg=lambda_reg)
            residuals_iter = y - X @ beta
            if robust:
                w = w * _tukey_weights(residuals=residuals_iter)

        if beta is None:
            beta = _weighted_ridge_solve(X=X, y=y, weights=w, lambda_reg=lambda_reg)

        level_bias = (X_full @ beta).reshape(full_shape)
        log_bias += level_bias
        residual = residual - level_bias

    # Center: ensure bias_field has unit mean within mask
    log_bias -= log_bias[mask].mean()
    return log_bias


def polynomial_bias_field_dwi(
    data,
    gtab,
    *,
    mask=None,
    order=3,
    pyramid_levels=(4, 2, 1),
    n_iter=4,
    lambda_reg=1e-3,
    robust=True,
    gradient_weighting=True,
    zero_background=False,
):
    """DWI bias field correction via multi-resolution Legendre polynomial regression.

    Estimates the bias field from the mean b0 volume in log space using
    coarse-to-fine Legendre polynomial regression, then applies the estimated
    field to all DWI volumes.

    Parameters
    ----------
    data : ndarray
        4D DWI data (X, Y, Z, N).
    gtab : GradientTable
        Gradient table.
    mask : ndarray, optional
        3D binary brain mask. Auto-computed via median_otsu if None.
    order : int, optional
        Maximum Legendre polynomial order (terms where i+j+k <= order).
    pyramid_levels : tuple of int, optional
        Downsampling factors for coarse-to-fine pyramid (descending order).
    n_iter : int, optional
        Reweighting iterations per pyramid level.
    lambda_reg : float, optional
        Ridge regularization strength.
    robust : bool, optional
        Apply Tukey biweight robust reweighting.
    gradient_weighting : bool, optional
        Apply gradient-based edge suppression.
    zero_background : bool, optional
        If True, set the bias field to 1.0 (no correction) outside the brain
        mask. If False, the raw extrapolated field values are preserved in
        the returned bias_field array. Has no effect on the corrected DWI
        data (background voxels are always zeroed by the brain mask).

    Returns
    -------
    corrected : ndarray
        Bias-corrected 4D DWI data, same dtype as input.
    bias_field : ndarray
        Estimated 3D multiplicative bias field.
    """
    orig_dtype = data.dtype
    mean_b0 = _get_mean_b0(data, gtab)
    mask = _get_mask(mean_b0, mask)
    log_b0 = np.log(np.clip(mean_b0, 1e-10, None))

    log_bias = _polynomial_pyramid_fit(
        log_b0=log_b0,
        mask=mask,
        order=order,
        pyramid_levels=pyramid_levels,
        n_iter=n_iter,
        lambda_reg=lambda_reg,
        robust=robust,
        gradient_weighting=gradient_weighting,
    )

    if zero_background:
        log_bias[~mask] = 0.0
    bias_field = np.exp(log_bias)
    corrected = applymask(data.astype(np.float64) / bias_field[..., None], mask).astype(
        orig_dtype
    )
    return corrected, bias_field


def _build_bspline_design_matrix_py(*, log_b0_shape, n_control, mask_flat):
    """Build sparse B-spline design matrix (pure Python fallback).

    Parameters
    ----------
    log_b0_shape : tuple of int
        Shape of the 3D volume (S, R, C).
    n_control : tuple of int
        Control grid dimensions (ns, nr, nc).
    mask_flat : ndarray
        Flattened boolean mask, shape (S*R*C,).

    Returns
    -------
    X : scipy.sparse.csr_matrix
        Design matrix, shape (N_masked, K_ctrl_total).
    """
    S, R, C = log_b0_shape
    ns, nr, nc = n_control
    K = ns * nr * nc

    mask_3d = mask_flat.reshape(log_b0_shape)
    iz_all, iy_all, ix_all = np.where(mask_3d)
    N = len(iz_all)

    def _vox_to_ctrl_arr(vox, shape_d, n_ctrl_d):
        if shape_d <= 1 or n_ctrl_d <= 1:
            return np.zeros(len(vox), dtype=np.float64)
        return vox.astype(np.float64) * (n_ctrl_d - 1) / (shape_d - 1)

    tz = _vox_to_ctrl_arr(iz_all, S, ns)
    ty = _vox_to_ctrl_arr(iy_all, R, nr)
    tx = _vox_to_ctrl_arr(ix_all, C, nc)

    def _bspline_basis_batch(t, n_ctrl):
        """Vectorized cubic B-spline basis.

        Returns (N,4) basis values and (N,4) control indices.
        """
        t = np.clip(t, 0.0, n_ctrl - 1 - 1e-10)
        k = np.floor(t).astype(np.int64)
        k = np.minimum(k, n_ctrl - 2)
        u = t - k
        u2 = u * u
        u3 = u2 * u
        b = np.stack(
            [
                (1.0 - u) ** 3 / 6.0,
                (3.0 * u3 - 6.0 * u2 + 4.0) / 6.0,
                (-3.0 * u3 + 3.0 * u2 + 3.0 * u + 1.0) / 6.0,
                u3 / 6.0,
            ],
            axis=-1,
        )  # (N, 4)
        ctrl = np.stack([k - 1, k, k + 1, k + 2], axis=-1)  # (N, 4)
        return b, ctrl

    bz, cz = _bspline_basis_batch(tz, ns)  # (N, 4)
    by_, cy = _bspline_basis_batch(ty, nr)
    bx, cx = _bspline_basis_batch(tx, nc)

    # Tensor product: (N, 4, 4, 4) via broadcasting
    vals = (
        bz[:, :, np.newaxis, np.newaxis]
        * by_[:, np.newaxis, :, np.newaxis]
        * bx[:, np.newaxis, np.newaxis, :]
    )
    cols = (
        cz[:, :, np.newaxis, np.newaxis] * (nr * nc)
        + cy[:, np.newaxis, :, np.newaxis] * nc
        + cx[:, np.newaxis, np.newaxis, :]
    )
    rows = np.broadcast_to(
        np.arange(N, dtype=np.int64)[:, np.newaxis, np.newaxis, np.newaxis],
        (N, 4, 4, 4),
    )

    # Validity: all three ctrl indices must be in bounds
    valid = (
        (cz[:, :, np.newaxis, np.newaxis] >= 0)
        & (cz[:, :, np.newaxis, np.newaxis] < ns)
        & (cy[:, np.newaxis, :, np.newaxis] >= 0)
        & (cy[:, np.newaxis, :, np.newaxis] < nr)
        & (cx[:, np.newaxis, np.newaxis, :] >= 0)
        & (cx[:, np.newaxis, np.newaxis, :] < nc)
    )

    return sparse.csr_matrix(
        (vals[valid], (rows[valid], cols[valid])),
        shape=(N, K),
        dtype=np.float64,
    )


def _build_bspline_design_matrix(*, log_b0_shape, n_control, mask_flat):
    """Build sparse B-spline design matrix, using Cython backend if available.

    Parameters
    ----------
    log_b0_shape : tuple of int
        Shape of the 3D volume (S, R, C).
    n_control : tuple of int
        Control grid dimensions (ns, nr, nc).
    mask_flat : ndarray
        Flattened boolean mask, shape (S*R*C,).

    Returns
    -------
    X : scipy.sparse.csr_matrix
        Design matrix, shape (N_masked, K_ctrl_total).
    """
    if _HAVE_CYTHON:
        S, R, C = log_b0_shape
        ns, nr, nc = n_control
        K = ns * nr * nc
        mask_3d = mask_flat.reshape(log_b0_shape).astype(np.uint8)
        N_max = int(mask_flat.sum())

        out_coords = np.zeros((N_max, 3), dtype=np.int64)
        N_actual = int(masked_voxel_coords(np.ascontiguousarray(mask_3d), out_coords))
        out_coords = out_coords[:N_actual]

        def _scale(axis_coords, shape_d, n_ctrl_d):
            if shape_d <= 1 or n_ctrl_d <= 1:
                return np.zeros(len(axis_coords), dtype=np.float64)
            return axis_coords.astype(np.float64) * (n_ctrl_d - 1) / (shape_d - 1)

        grid_coords = np.column_stack(
            [
                _scale(out_coords[:, 0], S, ns),
                _scale(out_coords[:, 1], R, nr),
                _scale(out_coords[:, 2], C, nc),
            ]
        ).astype(np.float64)

        n_ctrl_arr = np.array([ns, nr, nc], dtype=np.int64)
        row_ptr = np.zeros(N_actual + 1, dtype=np.int64)
        col_idx = np.zeros(N_actual * 64, dtype=np.int64)
        values = np.zeros(N_actual * 64, dtype=np.float64)

        nnz = int(
            evaluate_bspline_rows(
                np.ascontiguousarray(grid_coords),
                n_ctrl_arr,
                row_ptr,
                col_idx,
                values,
            )
        )
        col_idx = col_idx[:nnz]
        values = values[:nnz]

        return sparse.csr_matrix(
            (values, col_idx, row_ptr),
            shape=(N_actual, K),
            dtype=np.float64,
        )

    return _build_bspline_design_matrix_py(
        log_b0_shape=log_b0_shape, n_control=n_control, mask_flat=mask_flat
    )


def _sparse_weighted_ridge_solve(*, X_sparse, y, weights, lambda_reg):
    """Solve sparse weighted ridge regression.

    The Gram matrix A = X^T W X is computed via chunked dense BLAS DGEMM,
    which is substantially faster than scipy sparse×sparse multiplication
    when the result is nearly dense (K < 1000).

    Parameters
    ----------
    X_sparse : scipy.sparse.csr_matrix
        Design matrix, shape (N, K).
    y : ndarray
        Target values, shape (N,).
    weights : ndarray
        Non-negative regression weights, shape (N,).
    lambda_reg : float
        Ridge regularization strength.

    Returns
    -------
    beta : ndarray
        Coefficient vector, shape (K,).
    """
    K = X_sparse.shape[1]
    N = X_sparse.shape[0]

    if _HAVE_CYTHON:
        # Fast path: Cython direct CSR accumulation
        A = np.zeros((K, K), dtype=np.float64)
        b_vec = np.zeros(K, dtype=np.float64)
        gram_matrix_csr(
            np.asarray(X_sparse.data, dtype=np.float64),
            np.asarray(X_sparse.indices, dtype=np.int32),
            np.asarray(X_sparse.indptr, dtype=np.int32),
            np.ascontiguousarray(weights, dtype=np.float64),
            np.ascontiguousarray(y, dtype=np.float64),
            A,
            b_vec,
        )
    else:
        # Chunked BLAS DGEMM: avoids sparse×sparse which is slow for dense-ish
        # results (K×K), converting sparse rows to dense in blocks and using
        # BLAS for the accumulation.
        chunk = min(4096, N)
        A = np.zeros((K, K), dtype=np.float64)
        b_vec = np.zeros(K, dtype=np.float64)
        for i in range(0, N, chunk):
            Xc = X_sparse[i : i + chunk].toarray()  # (chunk, K)
            wc = weights[i : i + chunk]
            A += Xc.T @ (wc[:, np.newaxis] * Xc)  # BLAS DGEMM
            b_vec += Xc.T.dot(wc * y[i : i + chunk])

    A += lambda_reg * np.eye(K)
    try:
        beta = scipy_linalg.solve(A, b_vec, assume_a="pos")
    except scipy_linalg.LinAlgError:
        beta, _, _, _ = np.linalg.lstsq(A, b_vec, rcond=None)
    return beta


def _refine_control_coeffs(*, coeffs, n_ctrl_coarse, n_ctrl_fine):
    """Trilinear interpolation of control grid coefficients.

    Parameters
    ----------
    coeffs : ndarray
        Flattened control point coefficients at coarse resolution.
    n_ctrl_coarse : tuple of int
        Coarse control grid dimensions.
    n_ctrl_fine : tuple of int
        Fine control grid dimensions.

    Returns
    -------
    fine_coeffs : ndarray
        Flattened coefficients at fine resolution.
    """
    coarse_grid = coeffs.reshape(n_ctrl_coarse)
    zoom_factors = tuple(f / c for f, c in zip(n_ctrl_fine, n_ctrl_coarse))
    fine_grid = ndimage.zoom(coarse_grid, zoom=zoom_factors, order=1)
    # Trim or pad to exactly match target shape
    slices = tuple(slice(0, n) for n in n_ctrl_fine)
    fine_grid = fine_grid[slices]
    if fine_grid.shape != tuple(n_ctrl_fine):
        padded = np.zeros(n_ctrl_fine, dtype=np.float64)
        src_slices = tuple(slice(0, s) for s in fine_grid.shape)
        padded[src_slices] = fine_grid
        fine_grid = padded
    return fine_grid.ravel()


def _eval_bspline_field(*, coeffs, n_control, out_shape):
    """Evaluate B-spline field at all voxel positions.

    Uses ``scipy.ndimage.map_coordinates`` with ``prefilter=False`` so that
    ``coeffs`` are treated directly as B-spline weights (not as values to
    interpolate through).  This avoids building a full-resolution sparse
    design matrix and is O(N) in C rather than O(N × 64) in Python.

    Parameters
    ----------
    coeffs : ndarray
        Flattened control point coefficients (B-spline weights).
    n_control : tuple of int
        Control grid dimensions (ns, nr, nc).
    out_shape : tuple of int
        Output volume shape (S, R, C).

    Returns
    -------
    field : ndarray
        Evaluated field, shape out_shape.
    """
    S, R, C = out_shape
    ns, nr, nc = n_control

    coeff_grid = np.ascontiguousarray(coeffs.reshape(n_control), dtype=np.float64)

    iz = np.linspace(0, ns - 1, S) if ns > 1 else np.zeros(S)
    iy = np.linspace(0, nr - 1, R) if nr > 1 else np.zeros(R)
    ix_ = np.linspace(0, nc - 1, C) if nc > 1 else np.zeros(C)

    II, JJ, KK = np.meshgrid(iz, iy, ix_, indexing="ij")
    coords = np.vstack([II.ravel(), JJ.ravel(), KK.ravel()])

    field = ndimage.map_coordinates(
        coeff_grid, coords, order=3, mode="nearest", prefilter=False
    )
    return field.reshape(out_shape)


def _bspline_pyramid_fit(
    *,
    log_b0,
    mask,
    n_control_points,
    pyramid_levels,
    n_iter,
    lambda_reg,
    robust,
    gradient_weighting,
    sigma_factor=0.2,
):
    """Coarse-to-fine B-spline bias field regression.

    Parameters
    ----------
    log_b0 : ndarray
        3D log-domain b0 image, shape (S, R, C).
    mask : ndarray
        3D boolean brain mask.
    n_control_points : tuple of int
        Control grid dimensions at finest level.
    pyramid_levels : tuple of int
        Downsampling factors, ordered coarse-first (e.g. (4, 2, 1)).
    n_iter : int
        Reweighting iterations per pyramid level.
    lambda_reg : float
        Ridge regularization strength.
    robust : bool
        Apply Tukey biweight robust reweighting.
    gradient_weighting : bool
        Apply gradient-based edge suppression weights.
    sigma_factor : float, optional
        Sigma = factor * sigma_factor for Gaussian smoothing.

    Returns
    -------
    log_bias : ndarray
        Estimated log-domain bias field, same shape as log_b0.
    """
    full_shape = log_b0.shape

    if gradient_weighting:
        grad_w_full = _gradient_weights(log_b0=log_b0)

    # Center log_b0 so the B-spline fits only spatial variation, not the
    # DC offset (overall intensity level).
    log_b0_dc = log_b0[mask].mean()
    residual = log_b0 - log_b0_dc

    log_bias = np.zeros(full_shape, dtype=np.float64)
    prev_coeffs = None
    prev_n_ctrl = None

    for factor in pyramid_levels:
        n_ctrl = tuple(max(2, int(np.round(n / factor))) for n in n_control_points)

        if factor == 1:
            level_residual = residual
            level_mask = mask
        else:
            sigma = factor * sigma_factor
            smoothed = ndimage.gaussian_filter(residual, sigma=sigma)
            level_residual = ndimage.zoom(smoothed, zoom=1.0 / factor, order=1)
            level_mask = (
                ndimage.zoom(mask.astype(np.float64), zoom=1.0 / factor, order=0) > 0.5
            )

        level_shape = level_residual.shape
        mask_flat_level = level_mask.ravel()
        n_masked = mask_flat_level.sum()
        K = n_ctrl[0] * n_ctrl[1] * n_ctrl[2]

        if n_masked < K:
            continue

        y = level_residual.ravel()[mask_flat_level]

        X = _build_bspline_design_matrix(
            log_b0_shape=level_shape,
            n_control=n_ctrl,
            mask_flat=mask_flat_level,
        )

        # Warm-start: refine coefficients from previous coarser level
        if prev_coeffs is not None and prev_n_ctrl is not None:
            coeffs = _refine_control_coeffs(
                coeffs=prev_coeffs,
                n_ctrl_coarse=prev_n_ctrl,
                n_ctrl_fine=n_ctrl,
            )
        else:
            coeffs = np.zeros(K, dtype=np.float64)

        w = np.ones(n_masked, dtype=np.float64)
        if gradient_weighting:
            if factor == 1:
                gw = grad_w_full.ravel()[mask_flat_level]
            else:
                gw_down = ndimage.zoom(grad_w_full, zoom=1.0 / factor, order=1)
                gw = gw_down.ravel()[mask_flat_level]
            w = w * gw

        for _ in range(n_iter):
            coeffs = _sparse_weighted_ridge_solve(
                X_sparse=X, y=y, weights=w, lambda_reg=lambda_reg
            )
            residuals_iter = y - X @ coeffs
            if robust:
                w = w * _tukey_weights(residuals=residuals_iter)

        prev_coeffs = coeffs
        prev_n_ctrl = n_ctrl

        level_bias = _eval_bspline_field(
            coeffs=coeffs, n_control=n_ctrl, out_shape=full_shape
        )
        log_bias += level_bias
        residual = residual - level_bias

    # Center: ensure bias_field has unit mean within mask
    log_bias -= log_bias[mask].mean()
    return log_bias


def _auto_select_fit(
    *,
    log_b0,
    mean_b0,
    mask,
    order,
    n_control_points,
    pyramid_levels,
    n_iter,
    lambda_reg,
    robust,
    gradient_weighting,
):
    """Run poly and bspline fits, return the log-bias with lower CoV.

    Parameters
    ----------
    log_b0 : ndarray
        Log-domain mean b0, shape (X, Y, Z), float64.
    mean_b0 : ndarray
        Mean b0 in signal domain, shape (X, Y, Z), float64.
    mask : ndarray
        3D boolean brain mask.
    order : int
        Legendre polynomial order for poly fit.
    n_control_points : tuple of int
        B-spline control grid dimensions for bspline fit.
    pyramid_levels : tuple of int
        Downsampling factors for coarse-to-fine pyramid.
    n_iter : int
        Reweighting iterations per pyramid level.
    lambda_reg : float
        Ridge regularization strength.
    robust : bool
        Apply Tukey biweight robust reweighting.
    gradient_weighting : bool
        Apply gradient-based edge suppression.

    Returns
    -------
    log_bias : ndarray
        Log-domain bias field from the winning method.
    """
    log_bias_poly = _polynomial_pyramid_fit(
        log_b0=log_b0,
        mask=mask,
        order=order,
        pyramid_levels=pyramid_levels,
        n_iter=n_iter,
        lambda_reg=lambda_reg,
        robust=robust,
        gradient_weighting=gradient_weighting,
    )
    log_bias_bspline = _bspline_pyramid_fit(
        log_b0=log_b0,
        mask=mask,
        n_control_points=n_control_points,
        pyramid_levels=pyramid_levels,
        n_iter=n_iter,
        lambda_reg=lambda_reg,
        robust=robust,
        gradient_weighting=gradient_weighting,
    )

    def _cov(log_bf):
        """CoV of mean b0 corrected by the given log bias field."""
        corrected_b0 = mean_b0 / np.where(np.exp(log_bf) > 1e-10, np.exp(log_bf), 1.0)
        vals = corrected_b0[mask]
        return vals.std() / (vals.mean() + 1e-12)

    cov_poly = _cov(log_bias_poly)
    cov_bspline = _cov(log_bias_bspline)

    if cov_poly <= cov_bspline:
        logger.info(
            "bias_field_correction auto: selected 'poly' " "(CoV %.4f vs bspline %.4f)",
            cov_poly,
            cov_bspline,
        )
        return log_bias_poly

    logger.info(
        "bias_field_correction auto: selected 'bspline' " "(CoV %.4f vs poly %.4f)",
        cov_bspline,
        cov_poly,
    )
    return log_bias_bspline


def bias_field_correction(
    data,
    gtab,
    *,
    mask=None,
    method="bspline",
    order=3,
    n_control_points=(8, 8, 8),
    pyramid_levels=(4, 2, 1),
    n_iter=4,
    lambda_reg=1e-3,
    robust=True,
    gradient_weighting=True,
    return_bias_field=False,
    zero_background=False,
):
    """Top-level DWI bias field correction via regression.

    Estimates a smooth multiplicative bias field from the mean b0 volume
    using polynomial or B-spline regression in log space, then applies the
    correction uniformly to all DWI volumes.

    Parameters
    ----------
    data : ndarray
        4D DWI data (X, Y, Z, N).
    gtab : GradientTable
        Gradient table.
    mask : ndarray, optional
        3D binary brain mask. If None, computed via median_otsu.
    method : str, optional
        Bias correction method:

        - ``"poly"``: Legendre polynomial regression — fast, low-parameter.
        - ``"bspline"``: Cubic B-spline regression — more flexible.
        - ``"auto"``: Run both methods and return the one with lower
          Coefficient of Variation within the brain mask. The chosen method
          is logged at INFO level.
    order : int, optional
        Maximum Legendre polynomial degree (used only for method="poly").
    n_control_points : tuple of int, optional
        Control grid dimensions at finest level (used only for
        method="bspline").
    pyramid_levels : tuple of int, optional
        Downsampling factors for coarse-to-fine pyramid (descending order).
    n_iter : int, optional
        Reweighting iterations per pyramid level.
    lambda_reg : float, optional
        Ridge regularization strength.
    robust : bool, optional
        Apply Tukey biweight robust reweighting at each level.
    gradient_weighting : bool, optional
        Weight regression by edge-suppression map derived from the
        image gradient.
    return_bias_field : bool, optional
        If True, return the bias field alongside the corrected data.
    zero_background : bool, optional
        If True, set the bias field to 1.0 (no correction) outside the brain
        mask. If False, the raw extrapolated field values are preserved in
        the returned bias_field array. Has no effect on the corrected DWI
        data (background voxels are always zeroed by the brain mask).

    Returns
    -------
    corrected : ndarray
        Bias-corrected DWI, same dtype as input.
    bias_field : ndarray
        3D multiplicative bias field (only returned if
        return_bias_field=True).
    """
    orig_dtype = data.dtype
    mean_b0 = _get_mean_b0(data, gtab)
    mask = _get_mask(mean_b0, mask)
    log_b0 = np.log(np.clip(mean_b0.astype(np.float64), 1e-10, None))

    if method == "poly":
        log_bias = _polynomial_pyramid_fit(
            log_b0=log_b0,
            mask=mask,
            order=order,
            pyramid_levels=pyramid_levels,
            n_iter=n_iter,
            lambda_reg=lambda_reg,
            robust=robust,
            gradient_weighting=gradient_weighting,
        )
    elif method == "bspline":
        log_bias = _bspline_pyramid_fit(
            log_b0=log_b0,
            mask=mask,
            n_control_points=n_control_points,
            pyramid_levels=pyramid_levels,
            n_iter=n_iter,
            lambda_reg=lambda_reg,
            robust=robust,
            gradient_weighting=gradient_weighting,
        )
    elif method == "auto":
        log_bias = _auto_select_fit(
            log_b0=log_b0,
            mean_b0=mean_b0,
            mask=mask,
            order=order,
            n_control_points=n_control_points,
            pyramid_levels=pyramid_levels,
            n_iter=n_iter,
            lambda_reg=lambda_reg,
            robust=robust,
            gradient_weighting=gradient_weighting,
        )
    else:
        raise ValueError(f"method must be 'poly', 'bspline', or 'auto', got '{method}'")

    if zero_background:
        log_bias[~mask] = 0.0
    bias_field = np.exp(log_bias)
    corrected = applymask(data.astype(np.float64) / bias_field[..., None], mask).astype(
        orig_dtype
    )

    if return_bias_field:
        return corrected, bias_field
    return corrected



"""Tests for FORCE simulation module."""

import numpy as np
import pytest

from dipy.sims.force import (
    dispersion_lut,
    get_default_diffusivity_config,
    smallest_shell_bval,
    validate_diffusivity_config,
)


def test_dispersion_lut_structure():
    """Test dispersion_lut returns correct structure."""
    sphere = np.random.randn(10, 3)
    sphere = sphere / np.linalg.norm(sphere, axis=1, keepdims=True)
    odi_list = np.array([0.1, 0.2, 0.3])

    result = dispersion_lut(sphere, odi_list)

    assert isinstance(result, dict)
    assert len(result) == 10
    for i in range(10):
        assert i in result
        assert isinstance(result[i], dict)
        for odi in odi_list:
            assert odi in result[i]


def test_validate_diffusivity_config_valid():
    """Test validation of valid diffusivity config."""
    config = get_default_diffusivity_config()
    assert validate_diffusivity_config(config) is True


def test_validate_diffusivity_config_missing_key():
    """Test validation fails for missing keys."""
    config = {"wm_d_par_range": (2e-3, 3e-3)}
    with pytest.raises(ValueError, match="Missing required key"):
        validate_diffusivity_config(config)


def test_validate_diffusivity_config_invalid_range():
    """Test validation fails for invalid range."""
    config = get_default_diffusivity_config()
    config["wm_d_par_range"] = (3e-3, 2e-3)  # min > max
    with pytest.raises(ValueError, match="min must be <= max"):
        validate_diffusivity_config(config)


def test_get_default_diffusivity_config():
    """Test default config has all required keys."""
    config = get_default_diffusivity_config()

    assert "wm_d_par_range" in config
    assert "wm_d_perp_range" in config
    assert "gm_d_iso_range" in config
    assert "csf_d" in config

    # Check reasonable values
    assert config["csf_d"] > 0
    assert config["wm_d_par_range"][0] > 0
    assert config["wm_d_par_range"][1] >= config["wm_d_par_range"][0]


def test_smallest_shell_bval():
    """Test smallest shell b-value finding for n=1 (default)."""
    bvals = np.array([0, 0, 1000, 1000, 2000, 2000, 3000])

    min_shell, mask = smallest_shell_bval(bvals)

    assert len(min_shell) == 1
    assert min_shell[0] == 1000
    assert mask.sum() == 2
    assert mask[2] and mask[3]
    # Other shells must NOT be in the mask
    assert not mask[4] and not mask[5] and not mask[6]


def test_smallest_shell_bval_n2():
    """Test smallest_shell_bval returns two smallest shells with n=2."""
    bvals = np.array([0, 0, 1000, 1000, 2000, 2000, 3000, 3000])

    min_shells, mask = smallest_shell_bval(bvals, n=2)

    assert len(min_shells) == 2
    assert min_shells[0] == 1000
    assert min_shells[1] == 2000
    # b0s excluded from shell mask
    assert not mask[0] and not mask[1]
    # Both shells selected
    assert mask[2] and mask[3]  # b=1000
    assert mask[4] and mask[5]  # b=2000
    # Largest shell excluded
    assert not mask[6] and not mask[7]


def test_smallest_shell_bval_n_too_large():
    """Test smallest_shell_bval raises when n exceeds available shells."""
    bvals = np.array([0, 1000, 2000])
    with pytest.raises(ValueError, match="unique shells found"):
        smallest_shell_bval(bvals, n=3)


def test_smallest_shell_bval_no_nonzero():
    """Test smallest_shell_bval raises for all-zero bvals."""
    bvals = np.array([0, 0, 10, 20, 30])  # all below threshold
    with pytest.raises(ValueError, match="No non-b0 volumes"):
        smallest_shell_bval(bvals)


def test_save_load_force_simulations(tmp_path):
    """Test saving and loading FORCE simulations."""
    from dipy.sims.force import load_force_simulations, save_force_simulations

    # Create test simulations
    test_sims = {
        "signals": np.random.randn(10, 100).astype(np.float32),
        "labels": np.random.randint(0, 2, (10, 50)).astype(np.uint8),
        "fa": np.random.rand(10).astype(np.float32),
    }

    output_path = tmp_path / "test_sims.npz"
    save_force_simulations(test_sims, str(output_path))

    # Load and verify
    loaded = load_force_simulations(str(output_path))

    assert set(loaded.keys()) == set(test_sims.keys())
    for key in test_sims:
        np.testing.assert_array_equal(loaded[key], test_sims[key])


def _make_gtab(shells):
    """Build a minimal GradientTable with 6 isotropic directions per shell.

    Parameters
    ----------
    shells : list of int
        Non-zero b-values (one shell per entry). Two b0 volumes are prepended.
    """
    from dipy.core.gradients import gradient_table

    # 6 directions on a hemisphere (roughly isotropic)
    dirs = np.array(
        [
            [1, 0, 0],
            [-1, 0, 0],
            [0, 1, 0],
            [0, -1, 0],
            [0, 0, 1],
            [0, 0, -1],
        ],
        dtype=float,
    )
    dirs /= np.linalg.norm(dirs, axis=1, keepdims=True)

    bvals = [0, 0]
    bvecs = [[0, 0, 0], [0, 0, 0]]
    for b in shells:
        bvals.extend([b] * 6)
        bvecs.extend(dirs.tolist())

    return gradient_table(np.array(bvals, dtype=float), bvecs=np.array(bvecs))


def test_generate_force_simulations_compute_dti():
    """generate_force_simulations with compute_dti=True returns FA/MD/RD."""
    from dipy.sims.force import generate_force_simulations

    gtab = _make_gtab([1000])
    sims = generate_force_simulations(
        gtab,
        num_simulations=20,
        batch_size=20,
        num_cpus=1,
        compute_dti=True,
        compute_dki=False,
        verbose=False,
    )

    for key in ("fa", "md", "rd"):
        assert key in sims, f"Key '{key}' missing from simulations"
        arr = sims[key]
        assert arr.shape == (20,), f"Expected shape (20,) for '{key}', got {arr.shape}"
        assert np.any(arr != 0), f"'{key}' is all-zeros – DTI fitting appears skipped"

    # DKI keys must NOT be present
    for key in ("ak", "rk", "mk", "kfa"):
        assert key not in sims, f"DKI key '{key}' should be absent"


def test_generate_force_simulations_compute_dki():
    """generate_force_simulations with compute_dki=True returns AK/RK/MK/KFA."""
    from dipy.sims.force import generate_force_simulations

    # DKI requires at least 2 non-zero shells
    gtab = _make_gtab([1000, 2000])
    sims = generate_force_simulations(
        gtab,
        num_simulations=20,
        batch_size=20,
        num_cpus=1,
        compute_dti=True,
        compute_dki=True,
        verbose=False,
    )

    dki_keys = ("ak", "rk", "mk", "kfa")
    for key in dki_keys:
        assert key in sims, f"Key '{key}' missing from simulations"
        arr = sims[key]
        assert arr.shape == (20,), f"Expected shape (20,) for '{key}', got {arr.shape}"
        assert np.any(arr != 0), f"'{key}' is all-zeros – DKI fitting appears skipped"

    # DTI keys also present when compute_dti=True
    for key in ("fa", "md", "rd"):
        assert key in sims, f"DTI key '{key}' missing"


def test_generate_force_simulations_no_dti_no_dki():
    """generate_force_simulations with both flags False omits metric keys."""
    from dipy.sims.force import generate_force_simulations

    gtab = _make_gtab([1000])
    sims = generate_force_simulations(
        gtab,
        num_simulations=10,
        batch_size=10,
        num_cpus=1,
        compute_dti=False,
        compute_dki=False,
        verbose=False,
    )

    for key in ("fa", "md", "rd", "ak", "rk", "mk", "kfa"):
        if key in ("fa", "md", "rd"):
            # These keys exist but should be all zeros (initialised to zeros,
            # never filled when compute_dti=False)
            assert key in sims
            assert np.all(sims[key] == 0), f"'{key}' should be zero when DTI disabled"
        else:
            assert key not in sims, f"DKI key '{key}' should be absent"



import warnings

import numpy as np
from scipy.ndimage import label


def remove_holes_and_islands(
    binary_img, *, remove_holes=True, remove_islands=True, slice_wise=False
):
    """
    Remove any small mask chunks or holes
    that could be in the segmentation output.

    Parameters
    ----------
    binary_img : np.ndarray
        Binary image
    slice_wise : bool, optional
        Whether to run slice wise background correction as well
    remove_holes : bool, optional
        Whether to remove holes from the binary image
    remove_islands : bool, optional
        Whether to remove islands from the binary image

    Returns
    -------
    largest_img : np.ndarray
    """
    n_unique = len(np.unique(binary_img))
    if n_unique <= 1:
        return binary_img
    if n_unique > 2:
        warnings.warn(
            "The mask is not binary. \
                      Returning the original mask",
            UserWarning,
            stacklevel=2,
        )
        return binary_img

    if not remove_holes and not remove_islands:
        warnings.warn(
            "Both remove_holes and remove_islands is False. \
                      Returning the original mask",
            UserWarning,
            stacklevel=2,
        )
        return binary_img

    shape = binary_img.shape
    new_binary_img = binary_img.copy()

    if remove_islands:
        components, _ = label(new_binary_img)
        new_binary_img = components == np.argmax(np.bincount(components.flat)[1:]) + 1
        new_binary_img = new_binary_img.astype(int)
    if remove_holes:
        components, _ = label(1 - new_binary_img)
        new_binary_img = components == np.argmax(np.bincount(components.flat)[1:]) + 1
        new_binary_img = 1 - new_binary_img.astype(int)

    if slice_wise:
        for x in range(shape[0]):
            components, _ = label(1 - new_binary_img[x])
            temp = components == np.argmax(np.bincount(components.flat)[1:]) + 1
            new_binary_img[x] = 1 - temp.astype(int)
        for y in range(shape[1]):
            components, _ = label(1 - new_binary_img[:, y])
            temp = components == np.argmax(np.bincount(components.flat)[1:]) + 1
            new_binary_img[:, y] = 1 - temp.astype(int)
        for z in range(shape[2]):
            components, _ = label(1 - new_binary_img[..., z])
            temp = components == np.argmax(np.bincount(components.flat)[1:]) + 1
            new_binary_img[..., z] = 1 - temp.astype(int)

    return new_binary_img
